#!/usr/bin/env python3
"""
Test sentence segmentation against the "Golden Rules":
https://github.com/diasks2/pragmatic_segmenter#the-golden-rules
"""
import unittest

from gruut import sentences


class GoldenRulesTestCase(unittest.TestCase):
    """Test golden rules of sentence segmentation"""

    def test_rule_1(self):
        """Simple period to end sentence"""
        self.assertEqual(
            _get_sentences("Hello World. My name is Jonas."),
            ["Hello World.", "My name is Jonas."],
        )

    def test_rule_2(self):
        """Question mark to end sentence"""
        self.assertEqual(
            _get_sentences("What is your name? My name is Jonas."),
            ["What is your name?", "My name is Jonas."],
        )

    def test_rule_3(self):
        """Exclamation point to end sentence"""
        self.assertEqual(
            _get_sentences("There it is! I found it."), ["There it is!", "I found it."],
        )

    def test_rule_4(self):
        """One letter upper case abbreviations"""
        # NOTE: gruut removes the "." from E
        self.assertEqual(
            _get_sentences("My name is Jonas E. Smith."), ["My name is Jonas E Smith."],
        )

    def test_rule_5(self):
        """One letter lower case abbreviations"""
        # NOTE: gruut removes the "." from p
        self.assertEqual(
            _get_sentences("Please turn to p. 55."), ["Please turn to p 55."],
        )

    def test_rule_6(self):
        """Two letter lower case abbreviations in the middle of a sentence"""
        # NOTE: gruut expands abbreviations
        self.assertEqual(
            _get_sentences("Were Jane and co. at the party?"),
            ["Were Jane and company at the party?"],
        )

    def test_rule_7(self):
        """Two letter upper case abbreviations in the middle of a sentence"""
        # NOTE: gruut expands abbreviations
        self.assertEqual(
            _get_sentences("They closed the deal with Pitt, Briggs & Co. at noon."),
            ["They closed the deal with Pitt, Briggs and Company at noon."],
        )

    # def test_rule_8(self):
    #     """Two letter lower case abbreviations at the end of a sentence"""
    #     self.assertEqual(
    #         _get_sentences("Let's ask Jane and co. They should know."),
    #         ["Let's ask Jane and company. They should know."],
    #     )

    # def test_rule_9(self):
    #     """Two letter upper case abbreviations at the end of a sentence"""
    #     self.assertEqual(
    #         _get_sentences("They closed the deal with Pitt, Briggs & Co. It closed yesterday."),
    #         ["They closed the deal with Pitt, Briggs and Company. It closed yesterday."],
    #     )

    def test_rule_10(self):
        """Two letter (prepositive) abbreviations"""
        # NOTE: gruut expands abbreviations
        self.assertEqual(
            _get_sentences("I can see Mt. Fuji from here."),
            ["I can see Mount Fuji from here."],
        )

    # def test_rule_11(self):
    #     """Two letter (prepositive & postpositive) abbreviations"""
    #     self.assertEqual(
    #         _get_sentences("St. Michael's Church is on 5th st. near the light."),
    #         ["Saint Michael's Church is on 5th street near the light."],
    #     )

    def test_rule_12(self):
        """Possesive two letter abbreviations"""
        # NOTE: gruut expands abbreviations
        self.assertEqual(
            _get_sentences("That is JFK Jr.'s book."), ["That is J F K Junior's book."],
        )

    def test_rule_13(self):
        """Multi-period abbreviations in the middle of a sentence"""
        # NOTE: gruut expands abbreviations
        self.assertEqual(
            _get_sentences("I visited the U.S.A. last year."),
            ["I visited the U S A last year."],
        )

    # 14) Multi-period abbreviations at the end of a sentence
    # I live in the E.U. How about you?
    # ["I live in the E.U.", "How about you?"]

    # 15) U.S. as sentence boundary
    # I live in the U.S. How about you?
    # ["I live in the U.S.", "How about you?"]

    # 16) U.S. as non sentence boundary with next word capitalized
    # I work for the U.S. Government in Virginia.
    # ["I work for the U.S. Government in Virginia."]

    def test_rule_17(self):
        """U.S. as non sentence boundary"""
        # NOTE: gruut expands abbreviations
        self.assertEqual(
            _get_sentences("I have lived in the U.S. for 20 years."),
            ["I have lived in the U S for 20 years."],
        )

    # 18) A.M. / P.M. as non sentence boundary and sentence boundary
    # At 5 a.m. Mr. Smith went to the bank. He left the bank at 6 P.M. Mr. Smith then went to the store.
    # ["At 5 a.m. Mr. Smith went to the bank.", "He left the bank at 6 P.M.", "Mr. Smith then went to the store."]

    def test_rule_19(self):
        """Number as non sentence boundary"""
        self.assertEqual(
            _get_sentences("She has $100.00 in her bag."),
            ["She has $100.00 in her bag."],
        )

    # 21) Parenthetical inside sentence
    # He teaches science (He previously worked for 5 years as an engineer.) at the local University.
    # ["He teaches science (He previously worked for 5 years as an engineer.) at the local University."]

    # 22) Email addresses
    # Her email is Jane.Doe@example.com. I sent her an email.
    # ["Her email is Jane.Doe@example.com.", "I sent her an email."]

    # 23) Web addresses
    # The site is: https://www.example.50.com/new-site/awesome_content.html. Please check it out.
    # ["The site is: https://www.example.50.com/new-site/awesome_content.html.", "Please check it out."]

    # 24) Single quotations inside sentence
    # She turned to him, 'This is great.' she said.
    # ["She turned to him, 'This is great.' she said."]

    # 25) Double quotations inside sentence
    # She turned to him, "This is great." she said.
    # ["She turned to him, \"This is great.\" she said."]

    # 26) Double quotations at the end of a sentence
    # She turned to him, "This is great." She held the book out to show him.
    # ["She turned to him, \"This is great.\"", "She held the book out to show him."]

    def test_rule_27(self):
        """Double punctuation (exclamation point)"""
        self.assertEqual(
            _get_sentences("Hello!! Long time no see."),
            ["Hello!!", "Long time no see."],
        )

    def test_rule_28(self):
        """Double punctuation (question mark)"""
        self.assertEqual(
            _get_sentences("Hello?? Who is there?"), ["Hello??", "Who is there?"],
        )

    def test_rule_29(self):
        """Double punctuation (exclamation point / question mark)"""
        self.assertEqual(
            _get_sentences("Hello!? Is that you?"), ["Hello!?", "Is that you?"],
        )

    def test_rule_30(self):
        """Double punctuation (question mark / exclamation point)"""
        self.assertEqual(
            _get_sentences("Hello?! Is that you?"), ["Hello?!", "Is that you?"],
        )

    # 31) List (period followed by parens and no period to end item)
    # 1.) The first item 2.) The second item
    # ["1.) The first item", "2.) The second item"]

    # 32) List (period followed by parens and period to end item)
    # 1.) The first item. 2.) The second item.
    # ["1.) The first item.", "2.) The second item."]

    # 33) List (parens and no period to end item)
    # 1) The first item 2) The second item
    # ["1) The first item", "2) The second item"]

    # 34) List (parens and period to end item)
    # 1) The first item. 2) The second item.
    # ["1) The first item.", "2) The second item."]

    # 35) List (period to mark list and no period to end item)
    # 1. The first item 2. The second item
    # ["1. The first item", "2. The second item"]

    # 36) List (period to mark list and period to end item)
    # 1. The first item. 2. The second item.
    # ["1. The first item.", "2. The second item."]

    # 37) List with bullet
    # • 9. The first item • 10. The second item
    # ["• 9. The first item", "• 10. The second item"]

    # 38) List with hypthen
    # ⁃9. The first item ⁃10. The second item
    # ["⁃9. The first item", "⁃10. The second item"]

    # 39) Alphabetical list
    # a. The first item b. The second item c. The third list item
    # ["a. The first item", "b. The second item", "c. The third list item"]

    def test_rule_40(self):
        """Errant newlines in the middle of sentences (PDF)"""
        self.assertEqual(
            _get_sentences("This is a sentence\ncut off in the middle because pdf."),
            ["This is a sentence cut off in the middle because pdf."],
        )

    def test_rule_41(self):
        """Errant newlines in the middle of sentences"""
        self.assertEqual(
            _get_sentences("It was a cold \nnight in the city."),
            ["It was a cold night in the city."],
        )

    # 42) Lower case list separated by newline
    # features\ncontact manager\nevents, activities\n
    # ["features", "contact manager", "events, activities"]

    # 43) Geo Coordinates
    # You can find it at N°. 1026.253.553. That is where the treasure is.
    # ["You can find it at N°. 1026.253.553.", "That is where the treasure is."]

    # 44) Named entities with an exclamation point
    # She works at Yahoo! in the accounting department.
    # ["She works at Yahoo! in the accounting department."]

    # 45) I as a sentence boundary and I as an abbreviation
    # We make a good team, you and I. Did you see Albert I. Jones yesterday?
    # ["We make a good team, you and I.", "Did you see Albert I. Jones yesterday?"]

    # 46) Ellipsis at end of quotation
    # Thoreau argues that by simplifying one’s life, “the laws of the universe will appear less complex. . . .”
    # ["Thoreau argues that by simplifying one’s life, “the laws of the universe will appear less complex. . . .”"]

    # 47) Ellipsis with square brackets
    # "Bohr [...] used the analogy of parallel stairways [...]" (Smith 55).
    # ["\"Bohr [...] used the analogy of parallel stairways [...]\" (Smith 55)."]

    # 48) Ellipsis as sentence boundary (standard ellipsis rules)
    # If words are left off at the end of a sentence, and that is all that is omitted, indicate the omission with ellipsis marks (preceded and followed by a space) and then indicate the end of the sentence with a period . . . . Next sentence.
    # ["If words are left off at the end of a sentence, and that is all that is omitted, indicate the omission with ellipsis marks (preceded and followed by a space) and then indicate the end of the sentence with a period . . . .", "Next sentence."]

    # 49) Ellipsis as sentence boundary (non-standard ellipsis rules)
    # I never meant that.... She left the store.
    # ["I never meant that....", "She left the store."]

    # def test_rule_49(self):
    #     """Ellipsis as sentence boundary (non-standard ellipsis rules)"""
    #     self.assertEqual(
    #         _get_sentences("I never meant that.... She left the store."),
    #         ["I never meant that....", "She left the store."],
    #     )

    # 50) Ellipsis as non sentence boundary
    # I wasn’t really ... well, what I mean...see . . . what I'm saying, the thing is . . . I didn’t mean it.
    # ["I wasn’t really ... well, what I mean...see . . . what I'm saying, the thing is . . . I didn’t mean it."]

    # 51) 4-dot ellipsis
    # One further habit which was somewhat weakened . . . was that of combining words into self-interpreting compounds. . . . The practice was not abandoned. . . .
    # ["One further habit which was somewhat weakened . . . was that of combining words into self-interpreting compounds.", ". . . The practice was not abandoned. . . ."]


def _get_sentences(text):
    return [
        s.text
        for s in sentences(text, verbalize_numbers=False, verbalize_currency=False)
    ]


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Grapheme to phoneme prediction using python CRF suite.

Training requires pre-aligned corpus in Phonetisaurus format.
https://github.com/AdolfVonKleist/Phonetisaurus

The format of this corpus is:

    t}t e}ˈɛ s}s t}t

Each line contains a single word, with graphemes and phonemes separated by "}".
Multiple graphemes are separated by "|":

    s|h}ʃ o|w}ˈoʊ

The empty phoneme is "_":

    w}w h}_ y}ˈaɪ

Example:

.. code-block:: sh

    python3 -m gruut.g2p train --corpus g2p.corpus --output model.crf

Pre-trained models have the following settings:

* c1 = 0
* c2 = 1
* max-iterations = 100
"""
import argparse
import base64
import itertools
import logging
import os
import sys
import time
import typing
import unicodedata
from pathlib import Path

import pycrfsuite

_LOGGER = logging.getLogger("gruut.g2p")

# -----------------------------------------------------------------------------

FEATURES_TYPE = typing.Dict[str, typing.Union[str, bool, int, float]]
EPS_PHONEME = "_"
PHONEME_JOIN = "|"


class GraphemesToPhonemes:
    """Grapheme to phoneme CRF tagger"""

    def __init__(
        self,
        crf_tagger: typing.Union[str, Path, pycrfsuite.Tagger],
        eps_phoneme: str = EPS_PHONEME,
        phoneme_join: str = PHONEME_JOIN,
    ):
        if isinstance(crf_tagger, pycrfsuite.Tagger):
            self.crf_tagger = crf_tagger
        else:
            # Load model
            self.crf_tagger = pycrfsuite.Tagger()
            self.crf_tagger.open(str(crf_tagger))

        # Empty phoneme (dropped)
        self.eps_phoneme = eps_phoneme

        # String used to join multiple predicted phonemes
        self.phoneme_join = phoneme_join

    def __call__(self, word: str, normalize: bool = True) -> typing.Sequence[str]:
        """Guess phonemes for word"""
        features = GraphemesToPhonemes.word2features(word, normalize=normalize)
        coded_phonemes = self.crf_tagger.tag(features)
        phonemes: typing.List[str] = []

        for coded_ps in coded_phonemes:
            decoded_ps = GraphemesToPhonemes.decode_string(coded_ps)
            for p in decoded_ps.split(self.phoneme_join):
                if p != self.eps_phoneme:
                    phonemes.append(p)

        return phonemes

    # -------------------------------------------------------------------------

    @staticmethod
    def word2features(
        word: typing.Union[str, typing.List[str]], normalize: bool = True, **kwargs
    ):
        """Create feature dicts for all graphemes in a word"""
        if normalize and isinstance(word, str):
            # Combine characters
            # See: https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize
            word = unicodedata.normalize("NFC", word)

        return [
            GraphemesToPhonemes.grapheme2features(word, i, **kwargs)
            for i in range(len(word))
        ]

    @staticmethod
    def grapheme2features(
        word: typing.Union[str, typing.Sequence[str]],
        i: int,
        add_begin: bool = True,
        add_end: bool = True,
        chars_backward: int = 3,
        chars_forward: int = 3,
        bias: float = 1.0,
        encode: bool = True,
    ) -> FEATURES_TYPE:
        """Create feature dict for single grapheme"""
        g = word[i]
        num_g = len(word)

        features: FEATURES_TYPE = {
            "bias": bias,
            "grapheme": GraphemesToPhonemes.encode_string(g) if encode else g,
        }

        if (i == 0) and add_begin:
            features["begin"] = True

        for j in range(1, chars_backward + 1):
            if i >= j:
                g_prev = word[i - j]
                features[f"grapheme-{j}"] = (
                    GraphemesToPhonemes.encode_string(g_prev) if encode else g_prev
                )

        for j in range(1, chars_forward + 1):
            if i < (num_g - j):
                g_next = word[i + j]
                features[f"grapheme+{j}"] = (
                    GraphemesToPhonemes.encode_string(g_next) if encode else g_next
                )

        if (i == (num_g - 1)) and add_end:
            features["end"] = True

        return features

    @staticmethod
    def encode_string(s: str) -> str:
        """Encodes string in a form that crfsuite will accept (ASCII) and can be decoded"""
        return base64.b64encode(s.encode()).decode("ascii")

    @staticmethod
    def decode_string(s: str) -> str:
        """Decodes a string encoded by encode_string"""
        return base64.b64decode(s.encode("ascii")).decode()


# -----------------------------------------------------------------------------


def train(
    corpus_path: typing.Union[str, Path],
    output_path: typing.Union[str, Path],
    group_separator: str = "}",
    item_separator: str = "|",
    phoneme_join: str = PHONEME_JOIN,
    eps_phoneme: str = EPS_PHONEME,
    remove_phonemes: typing.Optional[typing.Iterable[str]] = None,
    c1: float = 0.0,
    c2: float = 1.0,
    max_iterations: int = 100,
):
    """Train a new G2P model"""
    corpus_path = Path(corpus_path)
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    remove_phonemes = set(remove_phonemes or [])

    trainer = pycrfsuite.Trainer(verbose=False)

    with open(corpus_path, "r", encoding="utf-8") as corpus:
        for i, line in enumerate(corpus):
            line = line.strip()
            if not line:
                continue

            # Parse into graphemes and phonemes
            skip_line = False
            parts = line.split()
            aligned_g = []
            aligned_p = []

            for part in parts:
                # Graphemes/phonemes are separated by }
                gs_str, ps_str = part.split(group_separator, maxsplit=1)

                # Multiple graphemes and phonemes are separated by |
                gs = gs_str.split(item_separator)
                ps = [
                    p for p in ps_str.split(item_separator) if p not in remove_phonemes
                ]

                # Align graphemes and phonemes, allowing for empty phonemes only
                for g1, p1 in itertools.zip_longest(gs, [ps], fillvalue=None):
                    if g1 is None:
                        skip_line = True
                        break

                    aligned_g.append(g1)

                    if p1:
                        aligned_p.append(phoneme_join.join(p1))
                    else:
                        aligned_p.append(eps_phoneme)

                if skip_line:
                    break

            if skip_line:
                _LOGGER.warning(
                    "Failed to align line %s: %s (graphemes=%s, phonemes=%s)",
                    i + 1,
                    line,
                    gs,
                    ps,
                )
                continue

            # Add example to trainer
            try:
                encoded_p = [GraphemesToPhonemes.encode_string(p) for p in aligned_p]
                trainer.append(GraphemesToPhonemes.word2features(aligned_g), encoded_p)
            except Exception as e:
                _LOGGER.exception("graphemes=%s phonemes=%s", aligned_g, aligned_p)
                raise e

    trainer.set_params(
        {
            "c1": c1,  # coefficient for L1 penalty
            "c2": c2,  # coefficient for L2 penalty
            "max_iterations": max_iterations,  # stop earlier
            # include transitions that are possible, but not observed
            "feature.possible_transitions": True,
        }
    )

    _LOGGER.debug(trainer.get_params())

    # Begin training
    _LOGGER.info("Training")

    start_time = time.perf_counter()
    trainer.train(str(output_path))
    end_time = time.perf_counter()

    _LOGGER.info("Training completed in %s second(s)", end_time - start_time)
    _LOGGER.info(trainer.logparser.last_iteration)


def do_train(args):
    """CLI method for train"""
    train(
        corpus_path=args.corpus,
        output_path=args.output,
        group_separator=args.group_separator,
        item_separator=args.item_separator,
        remove_phonemes=args.remove_phonemes,
        c1=args.c1,
        c2=args.c2,
        max_iterations=args.max_iterations,
    )


def do_predict(args):
    """CLI method for predict"""
    tagger = GraphemesToPhonemes(args.model)

    if args.texts:
        lines = args.texts
    else:
        lines = sys.stdin

        if os.isatty(sys.stdin.fileno()):
            print("Reading words from stdin...", file=sys.stderr)

    for line in lines:
        line = line.strip()
        if not line:
            continue

        word = line
        phonemes = tagger(word)

        print(word, *phonemes)


def do_test(args):
    """CLI method for test"""
    try:
        from rapidfuzz.string_metric import levenshtein
    except ImportError as e:
        _LOGGER.critical("rapidfuzz library is needed for levenshtein distance")
        _LOGGER.critical("pip install 'rapidfuzz>=1.4.1'")
        raise e

    tagger = GraphemesToPhonemes(args.model)

    # Load lexicon
    if args.texts:
        lines = args.texts
    else:
        lines = sys.stdin

        if os.isatty(sys.stdin.fileno()):
            print("Reading lexicon lines from stdin...", file=sys.stderr)

    lexicon = {}
    for line in lines:
        line = line.strip()
        if (not line) or (" " not in line):
            continue

        word, actual_phonemes = line.split(maxsplit=1)
        lexicon[word] = actual_phonemes

    # Predict phonemes
    predicted_phonemes = {}

    start_time = time.perf_counter()

    for word in lexicon:
        phonemes = tagger(word)
        predicted_phonemes[word] = " ".join(phonemes)

    end_time = time.perf_counter()

    # Calculate PER
    num_errors = 0
    num_missing = 0
    num_phonemes = 0

    for word, actual_phonemes in lexicon.items():
        expected_phonemes = predicted_phonemes.get(word, "")

        if expected_phonemes:
            distance = levenshtein(expected_phonemes, actual_phonemes)
            num_errors += distance
            num_phonemes += len(actual_phonemes)
        else:
            num_missing += 1
            _LOGGER.warning("No pronunciation for %s", word)

    assert num_phonemes > 0, "No phonemes were read"

    # Calculate results
    per = round(num_errors / num_phonemes, 2)
    wps = round(len(predicted_phonemes) / (end_time - start_time), 2)
    print("PER:", per, "Errors:", num_errors, "words/sec:", wps)

    if num_missing > 0:
        print("Total missing:", num_missing)


# -----------------------------------------------------------------------------


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(prog="g2p.py")

    # Create subparsers for each sub-command
    sub_parsers = parser.add_subparsers()
    sub_parsers.required = True
    sub_parsers.dest = "command"

    # -----
    # Train
    # -----
    train_parser = sub_parsers.add_parser(
        "train", help="Train a new G2P model from a pre-aligned Phonetisaurus corpus"
    )
    train_parser.add_argument(
        "--corpus", required=True, help="Path to aligned Phonetisaurus g2p corpus"
    )
    train_parser.add_argument(
        "--output", required=True, help="Path to output tagger model"
    )
    train_parser.add_argument("--c1", type=float, default=0.0, help="L1 penalty")
    train_parser.add_argument("--c2", type=float, default=1.0, help="L2 penalty")
    train_parser.add_argument(
        "--max-iterations",
        type=int,
        default=100,
        help="Maximum number of training iterations (default: 100)",
    )
    train_parser.add_argument(
        "--group-separator",
        default="}",
        help="Separator between graphemes and phonemes",
    )
    train_parser.add_argument(
        "--item-separator", default="|", help="Separator between items in a group"
    )
    train_parser.add_argument(
        "--remove-phonemes", nargs="*", help="Remove phonemes from examples"
    )
    train_parser.set_defaults(func=do_train)

    # -------
    # Predict
    # -------
    predict_parser = sub_parsers.add_parser(
        "predict", help="Predict phonemes for word(s)"
    )
    predict_parser.add_argument(
        "--model", required=True, help="Path to G2P tagger model"
    )
    predict_parser.add_argument("texts", nargs="*", help="Words")
    predict_parser.set_defaults(func=do_predict)

    # ----
    # Test
    # ----
    test_parser = sub_parsers.add_parser("test", help="Test G2P model on a lexicon")
    test_parser.add_argument("--model", required=True, help="Path to G2P tagger model")
    test_parser.add_argument(
        "texts", nargs="*", help="Lines with '<word> <phoneme> <phoneme> ...'"
    )
    test_parser.set_defaults(func=do_test)

    # ----------------
    # Shared arguments
    # ----------------
    for sub_parser in [train_parser, predict_parser, test_parser]:
        sub_parser.add_argument(
            "--debug", action="store_true", help="Print DEBUG messages to console"
        )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    args.func(args)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Tokenizes, verbalizes, and phonemizes text and SSML"""
import itertools
import logging
import re
import typing
import xml.etree.ElementTree as etree
from decimal import Decimal
from pathlib import Path

import babel
import babel.numbers
import dateparser
import networkx as nx
from gruut_ipa import IPA
from num2words import num2words

from gruut.const import (
    DATA_PROP,
    PHONEMES_TYPE,
    REGEX_PATTERN,
    BreakNode,
    BreakType,
    BreakWordNode,
    EndElement,
    GraphType,
    IgnoreNode,
    InlineLexicon,
    InterpretAs,
    InterpretAsFormat,
    Lexeme,
    MarkNode,
    Node,
    ParagraphNode,
    PunctuationWordNode,
    Sentence,
    SentenceNode,
    SpeakNode,
    SSMLParsingState,
    TextProcessorSettings,
    Word,
    WordNode,
    WordRole,
)
from gruut.lang import get_settings
from gruut.utils import (
    attrib_no_namespace,
    leaves,
    load_lexicon,
    maybe_split_ipa,
    pipeline_split,
    pipeline_transform,
    resolve_lang,
    tag_no_namespace,
    text_and_elements,
)

# -----------------------------------------------------------------------------

_LOGGER = logging.getLogger("gruut.text_processor")

DEFAULT_LEXICON_ID = ""


# -----------------------------------------------------------------------------


class TextProcessor:
    """Tokenizes, verbalizes, and phonemizes text and SSML"""

    def __init__(
        self,
        default_lang: str = "en_US",
        model_prefix: str = "",
        lang_dirs: typing.Optional[typing.Dict[str, typing.Union[str, Path]]] = None,
        search_dirs: typing.Optional[typing.Iterable[typing.Union[str, Path]]] = None,
        settings: typing.Optional[
            typing.MutableMapping[str, TextProcessorSettings]
        ] = None,
        **kwargs,
    ):
        self.default_lang = default_lang
        self.default_settings_kwargs = kwargs

        self.model_prefix = model_prefix
        self.search_dirs = search_dirs

        if lang_dirs is None:
            lang_dirs = {}

        # Convert to Paths
        self.lang_dirs = {
            dir_lang: Path(dir_path) for dir_lang, dir_path in lang_dirs.items()
        }

        if settings is None:
            settings = {}

        self.settings = settings

    def sentences(
        self,
        graph: GraphType,
        root: Node,
        major_breaks: bool = True,
        minor_breaks: bool = True,
        punctuations: bool = True,
        explicit_lang: bool = True,
        phonemes: bool = True,
        break_phonemes: bool = True,
        pos: bool = True,
    ) -> typing.Iterable[Sentence]:
        """Processes text and returns each sentence"""

        def get_lang(lang: str) -> str:
            if explicit_lang or (lang != self.default_lang):
                return lang

            # Implicit default language
            return ""

        sentence: typing.Optional[Sentence] = None

        par_idx: int = -1
        sent_idx: int = 0

        sent_pause_before_ms: int = 0
        sent_marks_before: typing.List[str] = []

        word_pause_before_ms: int = 0
        word_marks_before: typing.List[str] = []

        sentences: typing.List[Sentence] = []

        for dfs_node in nx.dfs_preorder_nodes(graph, root.node):
            node = graph.nodes[dfs_node][DATA_PROP]
            if isinstance(node, ParagraphNode):
                par_idx += 1
                sent_idx = 0
            elif isinstance(node, SentenceNode):
                # New sentence
                sentences.append(
                    Sentence(
                        idx=sent_idx,
                        par_idx=par_idx,
                        text="",
                        text_with_ws="",
                        text_spoken="",
                        voice=node.voice,
                        lang=get_lang(node.lang),
                        pause_before_ms=sent_pause_before_ms,
                        marks_before=(sent_marks_before if sent_marks_before else None),
                    )
                )

                sent_pause_before_ms = 0
                sent_marks_before = []
                sent_idx += 1
            elif graph.out_degree(dfs_node) == 0:
                if isinstance(node, WordNode):
                    assert sentences, "No sentence"
                    sentence = sentences[-1]

                    word_node = typing.cast(WordNode, node)
                    sentence.words.append(
                        Word(
                            idx=len(sentence.words),
                            sent_idx=sentence.idx,
                            par_idx=sentence.par_idx,
                            text=word_node.text,
                            text_with_ws=word_node.text_with_ws,
                            phonemes=word_node.phonemes if phonemes else None,
                            pos=word_node.pos if pos else None,
                            lang=get_lang(node.lang),
                            voice=node.voice,
                            pause_before_ms=word_pause_before_ms,
                            marks_before=(
                                word_marks_before if word_marks_before else None
                            ),
                        )
                    )

                    word_pause_before_ms = 0
                    word_marks_before = []
                elif isinstance(node, BreakWordNode):
                    assert sentences, "No sentence"
                    sentence = sentences[-1]

                    break_word_node = typing.cast(BreakWordNode, node)
                    is_minor_break = break_word_node.break_type == BreakType.MINOR
                    is_major_break = break_word_node.break_type == BreakType.MAJOR

                    if (minor_breaks and is_minor_break) or (
                        major_breaks and is_major_break
                    ):
                        sentence.words.append(
                            Word(
                                idx=len(sentence.words),
                                sent_idx=sentence.idx,
                                par_idx=sentence.par_idx,
                                text=break_word_node.text,
                                text_with_ws=break_word_node.text_with_ws,
                                phonemes=self._phonemes_for_break(
                                    break_word_node.break_type,
                                    lang=break_word_node.lang,
                                )
                                if phonemes and break_phonemes
                                else None,
                                is_minor_break=is_minor_break,
                                is_major_break=is_major_break,
                                lang=get_lang(node.lang),
                                voice=node.voice,
                                pause_before_ms=word_pause_before_ms,
                                marks_before=(
                                    word_marks_before if word_marks_before else None
                                ),
                            )
                        )

                        word_pause_before_ms = 0
                        word_marks_before = []
                elif punctuations and isinstance(node, PunctuationWordNode):
                    assert sentences, "No sentence"
                    sentence = sentences[-1]

                    punct_word_node = typing.cast(PunctuationWordNode, node)
                    sentence.words.append(
                        Word(
                            idx=len(sentence.words),
                            sent_idx=sentence.idx,
                            par_idx=sentence.par_idx,
                            text=punct_word_node.text,
                            text_with_ws=punct_word_node.text_with_ws,
                            is_punctuation=True,
                            lang=get_lang(punct_word_node.lang),
                            pause_before_ms=word_pause_before_ms,
                            marks_before=(
                                word_marks_before if word_marks_before else None
                            ),
                        )
                    )

                    word_pause_before_ms = 0
                    word_marks_before = []
                elif isinstance(node, BreakNode):
                    # Pause for some time
                    break_node = typing.cast(BreakNode, node)
                    break_parent = self._find_parent(
                        graph, node, (SentenceNode, ParagraphNode, SpeakNode)
                    )

                    if break_parent is not None:
                        break_ms = break_node.get_milliseconds()
                        break_parent_edges = list(graph.out_edges(break_parent.node))
                        break_edge_idx = break_parent_edges.index(
                            (break_parent.node, break_node.node)
                        )
                        is_last_edge = break_edge_idx == (len(break_parent_edges) - 1)

                        if isinstance(break_parent, SentenceNode):
                            assert sentences
                            sentence = sentences[-1]
                            if is_last_edge:
                                # End of sentence, add pause after
                                sentence.pause_after_ms += break_ms
                            elif sentence.words:
                                # Between words, add pause after previous word
                                sentence.words[-1].pause_after_ms += break_ms
                            else:
                                # Before first word, set pause for first word
                                word_pause_before_ms += break_ms
                        elif isinstance(break_parent, ParagraphNode):
                            if sentences and (sentences[-1].par_idx == par_idx):
                                # Between sentences in the same paragraph, add pause after previous sentence
                                sentences[-1].pause_after_ms += break_ms
                            else:
                                # Add pause to beginning of next sentence
                                sent_pause_before_ms += break_ms
                        elif isinstance(break_parent, SpeakNode):
                            if sentences:
                                # After paragraphs or sentences
                                sentences[-1].pause_after_ms += break_ms
                            else:
                                # Before any paragraphs or sentences
                                sent_pause_before_ms += break_ms
                elif isinstance(node, MarkNode):
                    # User-defined mark
                    mark_node = typing.cast(MarkNode, node)
                    mark_name = mark_node.name
                    mark_parent = self._find_parent(
                        graph, node, (SentenceNode, ParagraphNode, SpeakNode)
                    )

                    if mark_parent is not None:
                        mark_parent_edges = list(graph.out_edges(mark_parent.node))
                        mark_edge_idx = mark_parent_edges.index(
                            (mark_parent.node, mark_node.node)
                        )
                        is_last_edge = mark_edge_idx == (len(mark_parent_edges) - 1)

                        if isinstance(mark_parent, SentenceNode):
                            assert sentences
                            sentence = sentences[-1]
                            if is_last_edge:
                                # End of sentence, add mark after
                                if sentence.marks_after is None:
                                    sentence.marks_after = []

                                sentence.marks_after.append(mark_name)
                            elif sentence.words:
                                # Between words, add pause after previous word
                                last_word = sentence.words[-1]
                                if last_word.marks_after is None:
                                    last_word.marks_after = []

                                last_word.marks_after.append(mark_name)
                            else:
                                # Before first word, set pause for first word
                                word_marks_before.append(mark_name)
                        elif isinstance(mark_parent, ParagraphNode):
                            if sentences and (sentences[-1].par_idx == par_idx):
                                # Between sentences in the same paragraph, add pause after previous sentence
                                last_sentence = sentences[-1]
                                if last_sentence.marks_after is None:
                                    last_sentence.marks_after = []

                                last_sentence.marks_after.append(mark_name)
                            else:
                                # Add pause to beginning of next sentence
                                sent_marks_before.append(mark_name)
                        elif isinstance(mark_parent, SpeakNode):
                            if sentences:
                                # After paragraphs or sentences
                                last_sentence = sentences[-1]
                                if last_sentence.marks_after is None:
                                    last_sentence.marks_after = []

                                last_sentence.marks_after.append(mark_name)
                            else:
                                # Before any paragraphs or sentences
                                sent_marks_before.append(mark_name)

        # Post-process sentences to fix up text, voice, etc.
        for sentence in sentences:
            settings = self.get_settings(sentence.lang)
            if settings.keep_whitespace:
                # Whitespace is preseved
                sentence.text_with_ws = "".join(w.text_with_ws for w in sentence.words)
            else:
                # Make a best guess.
                # The join string is used before spoken words (except the first word).
                # This should have the effect of keeping punctuation next to words.
                word_texts: typing.List[str] = []
                for word in sentence.words:
                    if word.is_spoken:
                        if word_texts:
                            word_texts.append(f"{settings.join_str}{word.text}")
                        else:
                            word_texts.append(word.text)
                    else:
                        word_texts.append(word.text)

                sentence.text_with_ws = "".join(word_texts)

            sentence.text = settings.normalize_whitespace(sentence.text_with_ws)
            sentence.text_spoken = settings.join_str.join(
                w.text for w in sentence.words if w.is_spoken
            )

            # Normalize voice
            sent_voice = sentence.voice

            # Get voice used across all words
            for word in sentence.words:
                if word.voice:
                    if sent_voice and (sent_voice != word.voice):
                        # Multiple voices
                        sent_voice = ""
                        break

                    sent_voice = word.voice

            if sent_voice:
                sentence.voice = sent_voice

                # Set voice on all words
                for word in sentence.words:
                    word.voice = sent_voice

        return sentences

    def words(self, graph: GraphType, root: Node, **kwargs) -> typing.Iterable[Word]:
        """Processes text and returns each word"""
        for sent in self.sentences(graph, root, **kwargs):
            for word in sent:
                yield word

    def get_settings(self, lang: typing.Optional[str] = None) -> TextProcessorSettings:
        """Gets or creates settings for a language"""
        lang = lang or self.default_lang
        lang_settings = self.settings.get(lang)

        if lang_settings is not None:
            return lang_settings

        # Try again with resolved language
        resolved_lang = resolve_lang(lang)
        lang_settings = self.settings.get(resolved_lang)
        if lang_settings is not None:
            # Patch for the future
            self.settings[lang] = self.settings[resolved_lang]
            return lang_settings

        _LOGGER.debug(
            "No custom settings for language %s (%s). Creating default settings.",
            lang,
            resolved_lang,
        )

        # Create default settings for language
        lang_dir = self.lang_dirs.get(lang)
        lang_settings = get_settings(
            lang,
            lang_dir=lang_dir,
            model_prefix=self.model_prefix,
            search_dirs=self.search_dirs,
            **self.default_settings_kwargs,
        )
        self.settings[lang] = lang_settings
        self.settings[resolved_lang] = lang_settings

        return lang_settings

    # -------------------------------------------------------------------------
    # Processing
    # -------------------------------------------------------------------------

    def __call__(self, *args, **kwargs):
        """Processes text or SSML"""
        return self.process(*args, **kwargs)

    def process(
        self,
        text: str,
        lang: typing.Optional[str] = None,
        ssml: bool = False,
        pos: bool = True,
        phonemize: bool = True,
        post_process: bool = True,
        add_speak_tag: bool = True,
        detect_numbers: bool = True,
        detect_currency: bool = True,
        detect_dates: bool = True,
        detect_times: bool = True,
        verbalize_numbers: bool = True,
        verbalize_currency: bool = True,
        verbalize_dates: bool = True,
        verbalize_times: bool = True,
        max_passes: int = 5,
    ) -> typing.Tuple[GraphType, Node]:
        """
        Processes text or SSML

        Args:
            text: input text or SSML (ssml=True)
            lang: default language of input text
            ssml: True if input text is SSML
            pos: False if part of speech tagging should be disabled
            phonemize: False if phonemization should be disabled
            post_process: False if sentence/graph post-processing should be disabled
            add_speak_tag: True if <speak> should be automatically added to input text when ssml=True
            detect_numbers: True if numbers should be annotated in text (interpret_as="number")
            detect_currency: True if currency amounts should be annotated in text (interpret_as="currency")
            detect_dates: True if dates should be annotated in text (interpret_as="date")
            detect_times: True if clock times should be annotated in text (interpret_as="time")
            verbalize_numbers: True if annotated numbers should be expanded into words
            verbalize_currency: True if annotated currency amounts should be expanded into words
            verbalize_dates: True if annotated dates should be expanded into words
            verbalize_times: True if annotated clock times should be expanded into words

        Returns:
            graph, root: text graph and root node

        """
        if ssml:
            try:
                root_element = etree.fromstring(text)
            except Exception as e:
                if add_speak_tag:
                    # Try wrapping text in <speak> and parsing again
                    root_element = etree.fromstring(f"<speak>{text}</speak>")
                else:
                    # Log and re-raise exception
                    _LOGGER.exception("TextProcessor.process")
                    raise e

            def iter_elements():
                yield from text_and_elements(root_element)

        else:
            # Not XML
            def iter_elements():
                yield text

        graph = typing.cast(GraphType, nx.DiGraph())

        # Parse XML
        last_paragraph: typing.Optional[ParagraphNode] = None
        last_sentence: typing.Optional[SentenceNode] = None
        last_speak: typing.Optional[SpeakNode] = None
        root: typing.Optional[SpeakNode] = None
        parsing_state = SSMLParsingState.DEFAULT

        # [voice]
        voice_stack: typing.List[str] = []

        # [(interpret_as, format)]
        say_as_stack: typing.List[typing.Tuple[str, str]] = []

        # [(tag, lang)]
        lang_stack: typing.List[typing.Tuple[str, str]] = []
        current_lang: str = lang or self.default_lang

        # [lexicon.id]
        lookup_stack: typing.List[str] = []
        lexicon_id: typing.Optional[str] = None
        lexeme: typing.Optional[Lexeme] = None

        # id -> lexicon
        inline_lexicons: typing.Dict[str, InlineLexicon] = {}

        # True if current word is the last one
        is_last_word: bool = False

        # Current word's role
        word_role: typing.Optional[str] = None

        # Alias from <sub>
        last_alias: typing.Optional[str] = None

        # Used to skip <metadata>
        skip_elements: bool = False

        # Phonemes to use for next word(s)
        word_phonemes: typing.Optional[typing.List[typing.List[str]]] = None

        # Create __init__ args for new Node
        def scope_kwargs(target_class):
            scope = {}
            if voice_stack:
                scope["voice"] = voice_stack[-1]

            scope["lang"] = current_lang

            if target_class is WordNode:
                if say_as_stack:
                    scope["interpret_as"], scope["format"] = say_as_stack[-1]

                if word_role is not None:
                    scope["role"] = word_role

                if lookup_stack:
                    # Lexicon ids in order of look up
                    scope["lexicon_ids"] = list(reversed(lookup_stack))

            return scope

        def in_inline_lexicon(
            word_text: str, word_role: typing.Optional[str] = None
        ) -> bool:
            if inline_lexicons:
                for inline_lexicon_id in itertools.chain(
                    lookup_stack, [DEFAULT_LEXICON_ID]
                ):
                    maybe_lexicon = inline_lexicons.get(inline_lexicon_id)
                    if maybe_lexicon is None:
                        continue

                    maybe_role_phonemes = maybe_lexicon.words.get(word_text)
                    if maybe_role_phonemes is None:
                        continue

                    if (word_role is not None) and (word_role in maybe_role_phonemes):
                        # Role-specific pronunciation
                        return True

                    if WordRole.DEFAULT in maybe_role_phonemes:
                        # Default pronunciation
                        return True

            # No inline pronunciation
            return False

        # Process sub-elements and text chunks
        for elem_or_text in iter_elements():
            if isinstance(elem_or_text, str):
                if skip_elements:
                    # Inside <metadata>
                    continue

                # Text chunk
                text = typing.cast(str, elem_or_text)

                # <grapheme> inside <lexicon>
                if parsing_state == SSMLParsingState.IN_LEXICON_GRAPHEME:
                    assert lexeme is not None
                    lexeme.grapheme = text.strip()
                    continue

                # <phoneme> inside <lexicon>
                if parsing_state == SSMLParsingState.IN_LEXICON_PHONEME:
                    assert lexeme is not None
                    text = text.strip()

                    # Phonemes will be split on whitespace if at least one
                    # space is present, otherwise assume phonemes =
                    # graphemes.
                    lexeme.phonemes = maybe_split_ipa(text)
                    continue

                if last_alias is not None:
                    # Iniside a <sub>
                    text = last_alias

                if last_speak is None:
                    # Implicit <speak>
                    last_speak = SpeakNode(node=len(graph), implicit=True)
                    graph.add_node(last_speak.node, data=last_speak)
                    if root is None:
                        root = last_speak

                assert last_speak is not None

                if last_paragraph is None:
                    # Implicit <p>
                    p_node = ParagraphNode(
                        node=len(graph), implicit=True, **scope_kwargs(ParagraphNode)
                    )
                    graph.add_node(p_node.node, data=p_node)

                    graph.add_edge(last_speak.node, p_node.node)
                    last_paragraph = p_node

                assert last_paragraph is not None

                if last_sentence is None:
                    # Implicit <s>
                    s_node = SentenceNode(
                        node=len(graph), implicit=True, **scope_kwargs(SentenceNode)
                    )
                    graph.add_node(s_node.node, data=s_node)

                    graph.add_edge(last_paragraph.node, s_node.node)
                    last_sentence = s_node

                assert last_sentence is not None

                if parsing_state == SSMLParsingState.IN_WORD:
                    # No splitting
                    word_text = text
                    settings = self.get_settings(current_lang)
                    if (
                        settings.keep_whitespace
                        and (not is_last_word)
                        and (not word_text.endswith(settings.join_str))
                    ):
                        word_text += settings.join_str

                    word_kwargs = scope_kwargs(WordNode)
                    if word_phonemes:
                        word_kwargs["phonemes"] = word_phonemes.pop()

                    word_text_norm = settings.normalize_whitespace(word_text)

                    word_node = WordNode(
                        node=len(graph),
                        text=word_text_norm,
                        text_with_ws=word_text,
                        in_lexicon=(
                            in_inline_lexicon(word_text_norm, word_role)
                            or self._is_word_in_lexicon(word_text_norm, settings)
                        ),
                        **word_kwargs,
                    )
                    graph.add_node(word_node.node, data=word_node)
                    graph.add_edge(last_sentence.node, word_node.node)
                else:
                    # Split by whitespace
                    self._pipeline_tokenize(
                        graph,
                        last_sentence,
                        text,
                        word_phonemes=word_phonemes,
                        scope_kwargs=scope_kwargs(WordNode),
                        in_inline_lexicon=in_inline_lexicon,
                    )

            elif isinstance(elem_or_text, EndElement):
                # End of an element (e.g., </s>)
                end_elem = typing.cast(EndElement, elem_or_text)
                end_tag = tag_no_namespace(end_elem.element.tag)

                if end_tag == "voice":
                    if voice_stack:
                        voice_stack.pop()
                elif end_tag == "say-as":
                    if say_as_stack:
                        say_as_stack.pop()
                elif end_tag == "lookup":
                    if lookup_stack:
                        lookup_stack.pop()
                elif end_tag == "lexicon":
                    # Done parsing <lexicon>
                    parsing_state = SSMLParsingState.DEFAULT
                    lexicon_id = None
                elif (end_tag == "grapheme") and (
                    parsing_state == SSMLParsingState.IN_LEXICON_GRAPHEME
                ):
                    # Done with lexicon grapheme
                    parsing_state = SSMLParsingState.IN_LEXICON
                elif (end_tag == "phoneme") and (
                    parsing_state == SSMLParsingState.IN_LEXICON_PHONEME
                ):
                    # Done with lexicon phoneme
                    parsing_state = SSMLParsingState.IN_LEXICON
                elif (end_tag == "lexeme") and (
                    parsing_state == SSMLParsingState.IN_LEXICON
                ):
                    # Done with lexicon entry
                    assert lexeme is not None, "No lexeme"
                    assert (
                        lexeme.phonemes is not None
                    ), f"No phoneme for lexeme: {lexeme}"

                    assert lexicon_id is not None, "No lexicon id"
                    lexicon = inline_lexicons.get(lexicon_id)
                    assert lexicon is not None, f"No lexicon for id {lexicon_id}"

                    # Get or create role -> phonemes map
                    role_phonemes: typing.Dict[str, PHONEMES_TYPE] = lexicon.words.get(
                        lexeme.grapheme, {}
                    )

                    if lexeme.roles:
                        # Add phonemes for each role
                        for role in lexeme.roles:
                            role_phonemes[role] = lexeme.phonemes
                    else:
                        # Default (empty) role only
                        role_phonemes[WordRole.DEFAULT] = lexeme.phonemes

                    lexicon.words[lexeme.grapheme] = role_phonemes

                    # Reset state
                    lexeme = None
                else:
                    if lang_stack and (lang_stack[-1][0] == end_tag):
                        lang_stack.pop()

                    if lang_stack:
                        current_lang = lang_stack[-1][1]  # tag, lang
                    else:
                        current_lang = self.default_lang

                    if end_tag in {"w", "token"}:
                        # End of word
                        parsing_state = SSMLParsingState.DEFAULT
                        is_last_word = False
                        word_role = None
                    elif end_tag == "s":
                        # End of sentence
                        last_sentence = None
                    elif end_tag == "p":
                        # End of paragraph
                        last_paragraph = None
                    elif end_tag == "speak":
                        # End of speak
                        last_speak = root
                    elif end_tag == "sub":
                        # End of sub
                        last_alias = None
                    elif end_tag in {"metadata", "meta"}:
                        # End of metadata
                        skip_elements = False
                    elif end_tag == "phoneme":
                        # End of phoneme
                        word_phonemes = None
            else:
                if skip_elements:
                    # Inside <metadata>
                    continue

                # Start of an element (e.g., <p>)
                elem, elem_metadata = elem_or_text
                elem = typing.cast(etree.Element, elem)

                # Optional metadata for the element
                elem_metadata = typing.cast(
                    typing.Optional[typing.Dict[str, typing.Any]], elem_metadata
                )

                elem_tag = tag_no_namespace(elem.tag)

                if elem_tag == "speak":
                    # Explicit <speak>
                    maybe_lang = attrib_no_namespace(elem, "lang")
                    if maybe_lang:
                        lang_stack.append((elem_tag, maybe_lang))
                        current_lang = maybe_lang

                    speak_node = SpeakNode(
                        node=len(graph), element=elem, **scope_kwargs(SpeakNode)
                    )
                    if root is None:
                        root = speak_node

                    graph.add_node(speak_node.node, data=root)
                    last_speak = root
                elif elem_tag == "voice":
                    # Set voice scope
                    voice_name = attrib_no_namespace(elem, "name")
                    voice_stack.append(voice_name)
                elif elem_tag == "p":
                    # Explicit paragraph
                    if last_speak is None:
                        # Implicit <speak>
                        last_speak = SpeakNode(node=len(graph), implicit=True)
                        graph.add_node(last_speak.node, data=last_speak)
                        if root is None:
                            root = last_speak

                    assert last_speak is not None

                    maybe_lang = attrib_no_namespace(elem, "lang")
                    if maybe_lang:
                        lang_stack.append((elem_tag, maybe_lang))
                        current_lang = maybe_lang

                    p_node = ParagraphNode(
                        node=len(graph), element=elem, **scope_kwargs(ParagraphNode)
                    )
                    graph.add_node(p_node.node, data=p_node)
                    graph.add_edge(last_speak.node, p_node.node)
                    last_paragraph = p_node

                    # Force a new sentence to begin
                    last_sentence = None
                elif elem_tag == "s":
                    # Explicit sentence
                    if last_speak is None:
                        # Implicit <speak>
                        last_speak = SpeakNode(node=len(graph), implicit=True)
                        graph.add_node(last_speak.node, data=last_speak)
                        if root is None:
                            root = last_speak

                    assert last_speak is not None

                    if last_paragraph is None:
                        # Implicit paragraph
                        p_node = ParagraphNode(
                            node=len(graph), **scope_kwargs(ParagraphNode)
                        )
                        graph.add_node(p_node.node, data=p_node)

                        graph.add_edge(last_speak.node, p_node.node)
                        last_paragraph = p_node

                    maybe_lang = attrib_no_namespace(elem, "lang")
                    if maybe_lang:
                        lang_stack.append((elem_tag, maybe_lang))
                        current_lang = maybe_lang

                    s_node = SentenceNode(
                        node=len(graph), element=elem, **scope_kwargs(SentenceNode)
                    )
                    graph.add_node(s_node.node, data=s_node)
                    graph.add_edge(last_paragraph.node, s_node.node)
                    last_sentence = s_node
                elif elem_tag in {"w", "token"}:
                    # Explicit word
                    parsing_state = SSMLParsingState.IN_WORD
                    is_last_word = (
                        elem_metadata.get("is_last", False) if elem_metadata else False
                    )
                    maybe_lang = attrib_no_namespace(elem, "lang")
                    if maybe_lang:
                        lang_stack.append((elem_tag, maybe_lang))
                        current_lang = maybe_lang

                    word_role = attrib_no_namespace(elem, "role")
                elif elem_tag == "break":
                    # Break
                    last_target = last_sentence or last_paragraph or last_speak
                    assert last_target is not None
                    break_node = BreakNode(
                        node=len(graph),
                        element=elem,
                        time=attrib_no_namespace(elem, "time", ""),
                    )
                    graph.add_node(break_node.node, data=break_node)
                    graph.add_edge(last_target.node, break_node.node)
                elif elem_tag == "mark":
                    # Mark
                    last_target = last_sentence or last_paragraph or last_speak
                    assert last_target is not None
                    mark_node = MarkNode(
                        node=len(graph),
                        element=elem,
                        name=attrib_no_namespace(elem, "name", ""),
                    )
                    graph.add_node(mark_node.node, data=mark_node)
                    graph.add_edge(last_target.node, mark_node.node)
                elif elem_tag == "say-as":
                    say_as_stack.append(
                        (
                            attrib_no_namespace(elem, "interpret-as", ""),
                            attrib_no_namespace(elem, "format", ""),
                        )
                    )
                elif elem_tag == "sub":
                    # Sub
                    last_alias = attrib_no_namespace(elem, "alias", "")
                elif elem_tag in {"metadata", "meta"}:
                    # Metadata
                    skip_elements = True
                elif (elem_tag == "phoneme") and (
                    parsing_state != SSMLParsingState.IN_LEXICON
                ):
                    # Phonemes
                    word_phonemes_strs = attrib_no_namespace(elem, "ph", "").split()

                    if word_phonemes_strs:
                        # Phonemes will be split on whitespace if at least one
                        # space is present, otherwise assume phonemes =
                        # graphemes.
                        word_phonemes = [
                            maybe_split_ipa(phoneme_str)
                            for phoneme_str in word_phonemes_strs
                        ]
                    else:
                        word_phonemes = None
                elif elem_tag == "lang":
                    # Set language
                    maybe_lang = attrib_no_namespace(elem, "lang", "")
                    if maybe_lang:
                        lang_stack.append((elem_tag, maybe_lang))
                        current_lang = maybe_lang
                elif elem_tag == "lookup":
                    lookup_id = attrib_no_namespace(elem, "ref")
                    assert lookup_id is not None, f"Lookup id required ({elem})"
                    lookup_stack.append(lookup_id)
                elif elem_tag == "lexicon":
                    # Inline pronunciaton lexicon
                    # NOTE: Empty lexicon id means the "default" inline lexicon (<lookup> not required)
                    lexicon_id = attrib_no_namespace(elem, "id", DEFAULT_LEXICON_ID)
                    assert lexicon_id is not None

                    lexicon_alphabet = (
                        attrib_no_namespace(elem, "alphabet", "").strip().lower()
                    )
                    inline_lexicons[lexicon_id] = InlineLexicon(
                        lexicon_id=lexicon_id, alphabet=lexicon_alphabet
                    )

                    lexicon_uri = attrib_no_namespace(elem, "uri", "")
                    if lexicon_uri:
                        # Lexicon defined externally
                        _LOGGER.debug(
                            "Loading pronunciation lexicon from %s", lexicon_uri
                        )
                        load_lexicon(lexicon_uri, inline_lexicons[lexicon_id])
                    else:
                        # Lexicon defined within this document
                        parsing_state = SSMLParsingState.IN_LEXICON
                elif (elem_tag == "lexeme") and (
                    parsing_state == SSMLParsingState.IN_LEXICON
                ):
                    if lexeme is None:
                        lexeme = Lexeme()

                    role_str = attrib_no_namespace(elem, "role")
                    if role_str:
                        lexeme.roles = set(role_str.strip().split())
                elif (elem_tag == "grapheme") and (
                    parsing_state == SSMLParsingState.IN_LEXICON
                ):
                    # Inline pronunciaton lexicon (grapheme)
                    parsing_state = SSMLParsingState.IN_LEXICON_GRAPHEME
                    if lexeme is None:
                        lexeme = Lexeme()
                elif (elem_tag == "phoneme") and (
                    parsing_state == SSMLParsingState.IN_LEXICON
                ):
                    # Inline pronunciaton lexicon (phoneme)
                    parsing_state = SSMLParsingState.IN_LEXICON_PHONEME
                    if lexeme is None:
                        lexeme = Lexeme()

        assert root is not None

        # Do multiple passes over the graph
        num_passes_left = max_passes
        while num_passes_left > 0:
            was_changed = False

            # Do replacements before minor/major breaks
            if pipeline_split(self._split_replacements, graph, root):
                was_changed = True

            # Split punctuations (quotes, etc.) before breaks
            if pipeline_split(self._split_punctuations, graph, root):
                was_changed = True

            # Split on minor breaks (commas, etc.)
            if pipeline_split(self._split_minor_breaks, graph, root):
                was_changed = True

            # Expand abbrevations before major breaks
            if pipeline_split(self._split_abbreviations, graph, root):
                was_changed = True

            # Break apart initialisms (e.g., TTS or T.T.S.) before major breaks
            if pipeline_split(self._split_initialism, graph, root):
                was_changed = True

            # Split on major breaks (periods, etc.)
            if pipeline_split(self._split_major_breaks, graph, root):
                was_changed = True

            # Break apart sentences using BreakWordNodes
            if self._break_sentences(graph, root):
                was_changed = True

            # spell-out (e.g., abc -> a b c) before number expansion
            if pipeline_split(self._split_spell_out, graph, root):
                was_changed = True

            # Transform text into known classes.
            #
            # The order here is very important, since words with "interpret_as"
            # set will be skipped by later transformations.
            #
            # Dates are detected first so words like "1.1.2000" are not parsed
            # as numbers by Babel (the de_DE locale will parse this as 112000).
            #
            if detect_dates:
                if pipeline_transform(self._transform_date, graph, root):
                    was_changed = True

            if detect_currency:
                if pipeline_transform(self._transform_currency, graph, root):
                    was_changed = True

            if detect_numbers:
                if pipeline_transform(self._transform_number, graph, root):
                    was_changed = True

            if detect_times:
                if pipeline_transform(self._transform_time, graph, root):
                    was_changed = True

            # Verbalize known classes
            if verbalize_dates:
                if pipeline_transform(self._verbalize_date, graph, root):
                    was_changed = True

            if verbalize_times:
                if pipeline_transform(self._verbalize_time, graph, root):
                    was_changed = True

            if verbalize_numbers:
                if pipeline_transform(self._verbalize_number, graph, root):
                    was_changed = True

            if verbalize_currency:
                if pipeline_transform(self._verbalize_currency, graph, root):
                    was_changed = True

            # Break apart words
            if pipeline_split(self._break_words, graph, root):
                was_changed = True

            # Ignore non-words
            if pipeline_split(self._split_ignore_non_words, graph, root):
                was_changed = True

            if not was_changed:
                # No changes, so we can stop
                break

            num_passes_left -= 1

        # Gather words from leaves of the tree, group by sentence
        def process_sentence(words: typing.List[WordNode]):
            if pos:
                pos_settings = self.get_settings(node.lang)
                if pos_settings.get_parts_of_speech is not None:
                    pos_tags = pos_settings.get_parts_of_speech(
                        [word.text for word in words]
                    )
                    for word, pos_tag in zip(words, pos_tags):
                        word.pos = pos_tag

                        if not word.role:
                            word.role = f"gruut:{pos_tag}"

            if phonemize:
                # Add phonemes to word
                for word in words:
                    if word.phonemes:
                        # Word already has phonemes
                        continue

                    lexicon_ids: typing.List[str] = []

                    if word.lexicon_ids:
                        lexicon_ids.extend(word.lexicon_ids)

                    lexicon_ids.append(DEFAULT_LEXICON_ID)

                    # Look up phonemes from inline <lexicon>
                    for lexicon_id in lexicon_ids:
                        lexicon = inline_lexicons.get(lexicon_id)
                        if lexicon is None:
                            continue

                        maybe_role_phonemes = lexicon.words.get(word.text)
                        if maybe_role_phonemes is None:
                            continue

                        maybe_phonemes = maybe_role_phonemes.get(word.role)

                        if (maybe_phonemes is None) and (word.role != WordRole.DEFAULT):
                            # Try again with default role
                            maybe_phonemes = maybe_role_phonemes.get(WordRole.DEFAULT)

                        if maybe_phonemes is not None:
                            # Found inline pronunciation
                            word.phonemes = maybe_phonemes
                            break

                    if word.phonemes:
                        # Got phonemes from inline lexicon
                        continue

                    phonemize_settings = self.get_settings(word.lang)
                    if phonemize_settings.lookup_phonemes is not None:
                        word.phonemes = phonemize_settings.lookup_phonemes(
                            word.text, word.role
                        )

                    if (not word.phonemes) and (
                        phonemize_settings.guess_phonemes is not None
                    ):
                        word.phonemes = phonemize_settings.guess_phonemes(
                            word.text, word.role
                        )

        # Process tree leaves
        sentence_words: typing.List[WordNode] = []

        for dfs_node in nx.dfs_preorder_nodes(graph, root.node):
            node = graph.nodes[dfs_node][DATA_PROP]
            if isinstance(node, SentenceNode):
                if sentence_words:
                    process_sentence(sentence_words)
                    sentence_words = []
            elif graph.out_degree(dfs_node) == 0:
                if isinstance(node, WordNode):
                    word_node = typing.cast(WordNode, node)
                    sentence_words.append(word_node)

        if sentence_words:
            # Final sentence
            process_sentence(sentence_words)
            sentence_words = []

        if post_process:
            # Post-process sentences
            for dfs_node in nx.dfs_preorder_nodes(graph, root.node):
                node = graph.nodes[dfs_node][DATA_PROP]
                if isinstance(node, SentenceNode):
                    sent_node = typing.cast(SentenceNode, node)
                    sent_settings = self.get_settings(sent_node.lang)
                    if sent_settings.post_process_sentence is not None:
                        sent_settings.post_process_sentence(
                            graph, sent_node, sent_settings
                        )

            # Post process entire graph
            self.post_process_graph(graph, root)

        return graph, root

    def post_process_graph(self, graph: GraphType, root: Node):
        """User-defined post-processing of entire graph"""
        pass

    # -------------------------------------------------------------------------
    # Pipeline (custom)
    # -------------------------------------------------------------------------

    def _break_sentences(self, graph: GraphType, root: Node) -> bool:
        """Break sentences apart at BreakWordNode(break_type="major") nodes."""
        was_changed = False

        # This involves:
        # 1. Identifying where in the edge list of sentence the break occurs
        # 2. Creating a new sentence next to the existing one in the parent paragraph
        # 3. Moving everything after the break into the new sentence
        for leaf_node in list(leaves(graph, root)):
            if not isinstance(leaf_node, BreakWordNode):
                # Not a break
                continue

            break_word_node = typing.cast(BreakWordNode, leaf_node)
            if break_word_node.break_type != BreakType.MAJOR:
                # Not a major break
                continue

            # Get the path from the break up to the nearest sentence
            parent_node: int = next(iter(graph.predecessors(break_word_node.node)))
            parent: Node = graph.nodes[parent_node][DATA_PROP]
            s_path: typing.List[Node] = [parent]

            while not isinstance(parent, SentenceNode):
                parent_node = next(iter(graph.predecessors(parent_node)))
                parent = graph.nodes[parent_node][DATA_PROP]
                s_path.append(parent)

            # Should at least be [WordNode, SentenceNode]
            assert len(s_path) >= 2
            s_node = s_path[-1]
            assert isinstance(s_node, SentenceNode)

            if not s_node.implicit:
                # Don't break apart explicit sentences
                continue

            # Probably a WordNode
            below_s_node = s_path[-2]

            # Edges after the break will need to be moved to the new sentence
            s_edges = list(graph.out_edges(s_node.node))
            break_edge_idx = s_edges.index((s_node.node, below_s_node.node))

            edges_to_move = s_edges[break_edge_idx + 1 :]
            if not edges_to_move:
                # Final sentence, nothing to move
                continue

            # Locate parent paragraph so we can create a new sentence
            p_node = self._find_parent(graph, s_node, ParagraphNode)
            assert p_node is not None

            # Find the index of the edge between the paragraph and the current sentence
            p_s_edge = (p_node.node, s_node.node)
            p_edges = list(graph.out_edges(p_node.node))
            s_edge_idx = p_edges.index(p_s_edge)

            # Remove existing edges from the paragraph
            graph.remove_edges_from(p_edges)

            # Create a sentence and add an edge to it right after the current sentence
            new_s_node = SentenceNode(node=len(graph), implicit=True)
            graph.add_node(new_s_node.node, data=new_s_node)
            p_edges.insert(s_edge_idx + 1, (p_node.node, new_s_node.node))

            # Insert paragraph edges with new sentence
            graph.add_edges_from(p_edges)

            # Move edges from current sentence to new sentence
            graph.remove_edges_from(edges_to_move)
            graph.add_edges_from([(new_s_node.node, v) for (u, v) in edges_to_move])

            was_changed = True

        return was_changed

    def _break_words(self, graph: GraphType, node: Node):
        """Break apart words according to work breaks pattern"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as or word.in_lexicon or (not word.implicit):
            # Don't interpret words that are spoken for or explicit words (<w>)
            return

        settings = self.get_settings(word.lang)
        if settings.word_breaks_pattern is None:
            # No pattern set for this language
            return

        parts = settings.word_breaks_pattern.split(word.text)
        if len(parts) < 2:
            # Didn't split
            return

        # Preserve whitespace
        first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
        last_part_idx = len(parts) - 1

        for part_idx, part_text in enumerate(parts):
            part_text_norm = settings.normalize_whitespace(part_text)
            if not part_text_norm:
                continue

            if settings.keep_whitespace:
                if part_idx == 0:
                    part_text = first_ws + part_text

                if part_idx == last_part_idx:
                    part_text += last_ws
                else:
                    part_text += settings.join_str

            yield WordNode, {
                "text": part_text_norm,
                "text_with_ws": part_text,
                "implicit": True,
                "lang": word.lang,
                "voice": word.voice,
                "in_lexicon": self._is_word_in_lexicon(part_text_norm, settings),
                "is_from_broken_word": True,
            }

    def _split_punctuations(self, graph: GraphType, node: Node):
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as or word.in_lexicon:
            # Don't interpret words that are spoken for
            return

        settings = self.get_settings(word.lang)
        if (settings.begin_punctuations_pattern is None) and (
            settings.end_punctuations_pattern is None
        ):
            # No punctuation patterns
            return

        word_text = word.text
        first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
        has_punctuation = False

        # Punctuations at the beginning of the word
        if settings.begin_punctuations_pattern is not None:
            # Split into begin punctuation and rest of word
            parts = list(
                filter(
                    None,
                    settings.begin_punctuations_pattern.split(word_text, maxsplit=1),
                )
            )

            first_word = True
            while word_text and (len(parts) == 2):
                punct_text, word_text = parts
                if first_word:
                    # Preserve leadingwhitespace
                    punct_text = first_ws + punct_text
                    first_word = False

                punct_text_norm = settings.normalize_whitespace(punct_text)
                has_punctuation = True
                yield PunctuationWordNode, {
                    "text": punct_text_norm,
                    "text_with_ws": punct_text,
                    "implicit": True,
                    "lang": word.lang,
                    "voice": word.voice,
                }

                parts = list(
                    filter(
                        None,
                        settings.begin_punctuations_pattern.split(
                            word_text, maxsplit=1
                        ),
                    )
                )

        # Punctuations at the end of the word
        end_punctuations: typing.List[str] = []
        if settings.end_punctuations_pattern is not None:
            # Split into rest of word and end punctuation
            parts = list(
                filter(
                    None, settings.end_punctuations_pattern.split(word_text, maxsplit=1)
                )
            )

            while word_text and (len(parts) == 2):
                word_text, punct_text = parts
                has_punctuation = True
                end_punctuations.append(punct_text)
                parts = list(
                    filter(
                        None,
                        settings.end_punctuations_pattern.split(word_text, maxsplit=1),
                    )
                )

        if not has_punctuation:
            # Leave word as-is
            return

        if settings.keep_whitespace and (not end_punctuations):
            # Preserve trailing whitespace
            word_text = word_text + last_ws

        word_text_norm = settings.normalize_whitespace(word_text)

        if word_text:
            yield WordNode, {
                "text": word_text_norm,
                "text_with_ws": word_text,
                "implicit": True,
                "lang": word.lang,
                "voice": word.voice,
                "in_lexicon": self._is_word_in_lexicon(word_text_norm, settings),
            }

        last_punct_idx = len(end_punctuations) - 1
        for punct_idx, punct_text in enumerate(reversed(end_punctuations)):
            if settings.keep_whitespace and (punct_idx == last_punct_idx):
                # Preserve trailing whitespace
                punct_text += last_ws

            yield PunctuationWordNode, {
                "text": punct_text.strip(),
                "text_with_ws": punct_text,
                "implicit": True,
                "lang": word.lang,
                "voice": word.voice,
            }

    def _split_major_breaks(self, graph: GraphType, node: Node):
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as or word.in_lexicon:
            # Don't interpret words that are spoken for
            return

        settings = self.get_settings(word.lang)
        if settings.major_breaks_pattern is None:
            # No pattern set for this language
            return

        parts = settings.major_breaks_pattern.split(word.text_with_ws)
        if len(parts) < 2:
            return

        word_part = parts[0]
        break_part = parts[1]

        if word_part.strip():
            # Only yield word if there's anything but whitespace
            word_part_norm = settings.normalize_whitespace(word_part)

            yield WordNode, {
                "text": word_part_norm,
                "text_with_ws": word_part,
                "implicit": True,
                "lang": word.lang,
                "voice": word.voice,
                "in_lexicon": self._is_word_in_lexicon(word_part_norm, settings),
            }
        else:
            # Keep leading whitespace
            break_part = word_part + break_part

        yield BreakWordNode, {
            "break_type": BreakType.MAJOR,
            "text": settings.normalize_whitespace(break_part),
            "text_with_ws": break_part,
            "implicit": True,
            "lang": word.lang,
            "voice": word.voice,
        }

    def _split_minor_breaks(self, graph: GraphType, node: Node):
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as or word.in_lexicon:
            # Don't interpret words that are spoken for
            return

        settings = self.get_settings(word.lang)
        if settings.minor_breaks_pattern is None:
            # No pattern set for this language
            return

        parts = settings.minor_breaks_pattern.split(word.text_with_ws)
        if len(parts) < 2:
            return

        word_part = parts[0]
        if word_part.strip():
            # Only yield word if there's anything but whitespace
            word_part_norm = settings.normalize_whitespace(word_part)

            yield WordNode, {
                "text": word_part_norm,
                "text_with_ws": word_part,
                "implicit": True,
                "lang": word.lang,
                "voice": word.voice,
                "in_lexicon": self._is_word_in_lexicon(word_part_norm, settings),
            }

        break_part = parts[1]
        yield BreakWordNode, {
            "break_type": BreakType.MINOR,
            "text": settings.normalize_whitespace(break_part),
            "text_with_ws": break_part,
            "implicit": True,
            "lang": word.lang,
            "voice": word.voice,
        }

    def _find_parent(self, graph, node, *classes):
        """Tries to find a node whose type is in classes in the tree above node"""
        parents = []
        for parent_node in graph.predecessors(node.node):
            parent = graph.nodes[parent_node][DATA_PROP]
            if isinstance(parent, classes):
                return parent

            parents.append(parent)

        for parent in parents:
            match = self._find_parent(graph, parent, classes)
            if match is not None:
                return match

        return None

    # pylint: disable=no-self-use
    def _phonemes_for_break(
        self,
        break_type: typing.Union[str, BreakType],
        lang: typing.Optional[str] = None,
    ) -> typing.Optional[PHONEMES_TYPE]:
        if break_type == BreakType.MAJOR:
            return [IPA.BREAK_MAJOR.value]

        if break_type == BreakType.MINOR:
            return [IPA.BREAK_MINOR.value]

        return None

    # -------------------------------------------------------------------------

    def _pipeline_tokenize(
        self,
        graph,
        parent_node,
        text,
        word_phonemes: typing.Optional[typing.List[typing.List[str]]] = None,
        scope_kwargs=None,
        in_inline_lexicon: typing.Optional[
            typing.Callable[[str, typing.Optional[str]], bool]
        ] = None,
    ):
        """Splits text into word nodes"""
        if scope_kwargs is None:
            scope_kwargs = {}

        lang = self.default_lang
        if scope_kwargs is not None:
            lang = scope_kwargs.get("lang", lang)

        settings = self.get_settings(lang)
        assert settings is not None, f"No settings for {lang}"

        if settings.pre_process_text is not None:
            # Pre-process text
            text = settings.pre_process_text(text)

        # Split into separate words (preseving whitespace).
        for word_text in settings.split_words(text):
            word_text_norm = settings.normalize_whitespace(word_text)
            if not word_text_norm:
                continue

            if not settings.keep_whitespace:
                word_text = word_text_norm

            word_kwargs = scope_kwargs
            if word_phonemes:
                word_kwargs = {**scope_kwargs, "phonemes": word_phonemes.pop()}

            # Determine if word is in a lexicon.
            # If so, it will not be interpreted as an initialism, split apart, etc.
            in_lexicon: typing.Optional[bool] = None
            if in_inline_lexicon is not None:
                # Check inline <lexicon> first
                in_lexicon = in_inline_lexicon(
                    word_text_norm, scope_kwargs.get("word_role")
                )

            if not in_lexicon:
                # Check main language lexicon
                in_lexicon = self._is_word_in_lexicon(word_text_norm, settings)

            word_node = WordNode(
                node=len(graph),
                text=word_text_norm,
                text_with_ws=word_text,
                implicit=True,
                in_lexicon=in_lexicon,
                **word_kwargs,
            )
            graph.add_node(word_node.node, data=word_node)
            graph.add_edge(parent_node.node, word_node.node)

    # -------------------------------------------------------------------------
    # Pipeline Splits
    # -------------------------------------------------------------------------

    def _split_spell_out(self, graph: GraphType, node: Node):
        """Expand spell-out (a-1 -> a dash one)"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as != InterpretAs.SPELL_OUT:
            return

        settings = self.get_settings(word.lang)

        # Preserve whitespace
        first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
        last_char_idx = len(word.text) - 1

        for i, c in enumerate(word.text):
            # Look up in settings first ("." -> "dot")
            word_text = settings.spell_out_words.get(c)
            role = WordRole.DEFAULT

            if word_text is None:
                if c.isalpha():
                    # Assume this is a letter
                    word_text = c
                    role = WordRole.LETTER
                else:
                    # Leave as is (expand later in pipeline if digit, etc.)
                    word_text = c

            if not word_text:
                continue

            if settings.keep_whitespace:
                if i == 0:
                    word_text = first_ws + word_text

                if i == last_char_idx:
                    word_text += last_ws
                else:
                    word_text += settings.join_str

            yield WordNode, {
                "text": settings.normalize_whitespace(word_text),
                "text_with_ws": word_text,
                "implicit": True,
                "lang": word.lang,
                "role": role,
            }

    def _split_replacements(self, graph: GraphType, node: Node):
        """Do regex replacements on word text"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as or word.in_lexicon:
            # Don't interpret words that are spoken for
            return

        settings = self.get_settings(word.lang)

        if not settings.replacements:
            # No replacements
            return

        matched = False
        new_text = word.text_with_ws

        for pattern, template in settings.replacements:
            assert isinstance(pattern, REGEX_PATTERN)
            new_text, num_subs = pattern.subn(template, new_text)

            if num_subs > 0:
                matched = True

        if matched:
            # Tokenize new text (whitespace is preserved by regex)
            for part_text in settings.split_words(new_text):
                part_text_norm = settings.normalize_whitespace(part_text)

                if not settings.keep_whitespace:
                    part_text = part_text_norm

                if not part_text_norm:
                    # Ignore empty words
                    continue

                yield WordNode, {
                    "text": part_text_norm,
                    "text_with_ws": part_text,
                    "implicit": True,
                    "lang": word.lang,
                    "in_lexicon": self._is_word_in_lexicon(part_text_norm, settings),
                }

    def _split_abbreviations(self, graph: GraphType, node: Node):
        """Expand abbreviations"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as or word.in_lexicon:
            # Don't interpret words that are spoken for
            return

        settings = self.get_settings(word.lang)

        if not settings.abbreviations:
            # No abbreviations
            return

        new_text: typing.Optional[str] = None
        for pattern, template in settings.abbreviations.items():
            assert isinstance(pattern, REGEX_PATTERN), pattern
            match = pattern.match(word.text_with_ws)

            if match is not None:
                new_text = match.expand(template)
                break

        if new_text is not None:
            # Tokenize new text (whitespace should be preserved by regex)
            for part_text in settings.split_words(new_text):
                part_text_norm = settings.normalize_whitespace(part_text)
                if not part_text_norm:
                    continue

                if not settings.keep_whitespace:
                    part_text = part_text_norm

                yield WordNode, {
                    "text": part_text_norm,
                    "text_with_ws": part_text,
                    "implicit": True,
                    "lang": word.lang,
                    "in_lexicon": self._is_word_in_lexicon(part_text_norm, settings),
                }

    def _split_initialism(self, graph: GraphType, node: Node):
        """Split apart ABC or A.B.C."""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)

        if word.interpret_as or word.in_lexicon or (len(word.text) < 2):
            # Don't interpret words that are spoken for or are too short
            return

        settings = self.get_settings(word.lang)

        if (settings.is_initialism is None) or (settings.split_initialism is None):
            # Can't do anything without these functions
            return

        if not settings.is_initialism(word.text):
            # Not an initialism
            return

        first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
        parts = settings.split_initialism(word.text)
        last_part_idx = len(parts) - 1

        # Split according to language-specific function
        for part_idx, part_text in enumerate(parts):
            part_text_norm = settings.normalize_whitespace(part_text)
            if not part_text_norm:
                continue

            if settings.keep_whitespace:
                if part_idx == 0:
                    part_text = first_ws + part_text

                if 0 <= part_idx < last_part_idx:
                    part_text += settings.join_str
                elif part_idx == last_part_idx:
                    part_text += last_ws

            yield WordNode, {
                "text": part_text_norm,
                "text_with_ws": part_text,
                "implicit": True,
                "lang": word.lang,
                "role": WordRole.LETTER,
            }

    def _split_ignore_non_words(self, graph: GraphType, node: Node):
        """Mark non-words as ignored"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if word.interpret_as or word.in_lexicon:
            # Don't interpret words that are spoken for
            return

        settings = self.get_settings(word.lang)
        if settings.is_non_word is None:
            # No function for this language
            return

        if settings.is_non_word(word.text):
            yield (IgnoreNode, {})

    # -------------------------------------------------------------------------
    # Pipeline Transformations
    # -------------------------------------------------------------------------

    def _transform_number(self, graph: GraphType, node: Node) -> bool:
        if not isinstance(node, WordNode):
            return False

        word = typing.cast(WordNode, node)
        if (not word.is_maybe_number) or (
            word.interpret_as and (word.interpret_as != InterpretAs.NUMBER)
        ):
            return False

        settings = self.get_settings(word.lang)
        assert settings.babel_locale

        if settings.get_ordinal is not None:
            # Try to parse as an ordinal (e.g., 1st -> 1)
            ordinal_num = settings.get_ordinal(word.text)
            if ordinal_num is not None:
                word.interpret_as = InterpretAs.NUMBER
                word.format = InterpretAsFormat.NUMBER_ORDINAL
                word.number = Decimal(ordinal_num)
                return False

        try:
            # Try to parse as a number
            # This is important to handle thousand/decimal separators correctly.
            number = babel.numbers.parse_decimal(
                word.text, locale=settings.babel_locale
            )

            if not number.is_finite():
                raise ValueError("Not parsing nan or inf")

            word.interpret_as = InterpretAs.NUMBER

            if not word.format:
                # Retain ordinal, etc.
                word.format = InterpretAsFormat.NUMBER_CARDINAL

            word.number = number

            if (1000 < number < 3000) and (re.match(r"^\d+$", word.text) is not None):
                # Interpret numbers in this range as years by default, but only
                # if the text was entirely digits.
                #
                # So "2020" will become "twenty twenty", but "2,020" will become
                # "two thousand and twenty".
                word.format = InterpretAsFormat.NUMBER_YEAR
        except ValueError:
            # Probably not a number
            word.is_maybe_number = False

        return True

    def _transform_currency(self, graph: GraphType, node: Node,) -> bool:
        if not isinstance(node, WordNode):
            return False

        word = typing.cast(WordNode, node)
        if (not word.is_maybe_currency) or (
            word.interpret_as and (word.interpret_as != InterpretAs.CURRENCY)
        ):
            return False

        settings = self.get_settings(word.lang)

        if (settings.is_maybe_currency is not None) and (
            not settings.is_maybe_currency(word.text)
        ):
            # Probably not currency
            word.is_maybe_currency = False
            return False

        assert settings.babel_locale

        # Try to parse with known currency symbols
        parsed = False
        for currency_symbol in settings.currency_symbols:
            if word.text.startswith(currency_symbol):
                num_str = word.text[len(currency_symbol) :]
                try:
                    # Try to parse as a number
                    # This is important to handle thousand/decimal separators correctly.
                    number = babel.numbers.parse_decimal(
                        num_str, locale=settings.babel_locale
                    )
                    word.interpret_as = InterpretAs.CURRENCY
                    word.currency_symbol = currency_symbol
                    word.number = number
                    parsed = True
                    break
                except ValueError:
                    pass

        # If this *must* be a currency value, use the default currency
        if (not parsed) and (word.interpret_as == InterpretAs.CURRENCY):
            default_currency = settings.default_currency
            if default_currency:
                # Forced interpretation using default currency
                try:
                    number = babel.numbers.parse_decimal(
                        word.text, locale=settings.babel_locale
                    )
                    word.interpret_as = InterpretAs.CURRENCY
                    word.currency_name = default_currency
                    word.number = number
                except ValueError:
                    pass

        return True

    def _transform_date(self, graph: GraphType, node: Node):
        if not isinstance(node, WordNode):
            return False

        word = typing.cast(WordNode, node)
        if (not word.is_maybe_date) or (
            word.interpret_as and (word.interpret_as != InterpretAs.DATE)
        ):
            return False

        settings = self.get_settings(word.lang)

        try:
            if (settings.is_maybe_date is not None) and not settings.is_maybe_date(
                word.text
            ):
                # Probably not a date
                word.is_maybe_date = False
                return False

            assert settings.dateparser_lang

            dateparser_kwargs: typing.Dict[str, typing.Any] = {
                "settings": {"STRICT_PARSING": True},
                "languages": [settings.dateparser_lang],
            }

            date = dateparser.parse(word.text, **dateparser_kwargs)
            if date is not None:
                word.interpret_as = InterpretAs.DATE
                word.date = date
            elif word.interpret_as == InterpretAs.DATE:
                # Try again without strict parsing
                dateparser_kwargs["settings"]["STRICT_PARSING"] = False
                date = dateparser.parse(word.text, **dateparser_kwargs)
                if date is not None:
                    word.date = date
        except Exception:
            _LOGGER.exception("transform_date")

            # Not a date
            word.is_maybe_date = False
            return False

        return True

    def _transform_time(self, graph: GraphType, node: Node):
        if not isinstance(node, WordNode):
            return False

        word = typing.cast(WordNode, node)
        if (not word.is_maybe_time) or (
            word.interpret_as and (word.interpret_as != InterpretAs.TIME)
        ):
            return False

        settings = self.get_settings(word.lang)

        if settings.parse_time is None:
            # Can't parse a time anyways
            return False

        try:
            if (settings.is_maybe_time is not None) and not settings.is_maybe_time(
                word.text
            ):
                # Probably not a time
                word.is_maybe_time = False
                return False

            time = settings.parse_time(word.text)
            if time is not None:
                word.interpret_as = InterpretAs.TIME
                word.time = time
        except Exception:
            _LOGGER.exception("transform_time")

            # Not a time
            word.is_maybe_time = False
            return False

        return True

    def _is_word_in_lexicon(
        self, word: str, settings: TextProcessorSettings
    ) -> typing.Optional[bool]:
        """True if word is in the lexicon"""
        if settings.lookup_phonemes is None:
            return None

        return bool(settings.lookup_phonemes(word, do_transforms=False))

    # -------------------------------------------------------------------------
    # Verbalization
    # -------------------------------------------------------------------------

    def _verbalize_number(self, graph: GraphType, node: Node):
        """Split numbers into words"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if (word.interpret_as != InterpretAs.NUMBER) or (word.number is None):
            return

        settings = self.get_settings(word.lang)

        if (settings.is_maybe_number is not None) and not settings.is_maybe_number(
            word.text
        ):
            # Probably not a number
            return

        assert settings.num2words_lang
        num2words_kwargs = {"lang": settings.num2words_lang}
        decimal_nums = [word.number]

        if word.format == InterpretAsFormat.NUMBER_CARDINAL:
            num2words_kwargs["to"] = "cardinal"
        elif word.format == InterpretAsFormat.NUMBER_ORDINAL:
            num2words_kwargs["to"] = "ordinal"
        elif word.format == InterpretAsFormat.NUMBER_YEAR:
            num2words_kwargs["to"] = "year"
        elif word.format == InterpretAsFormat.NUMBER_DIGITS:
            num2words_kwargs["to"] = "cardinal"
            decimal_nums = [Decimal(d) for d in str(word.number.to_integral_value())]

        for decimal_num in decimal_nums:
            num_has_frac = (decimal_num % 1) != 0

            # num2words uses the number as an index sometimes, so it *has* to be
            # an integer, unless we're doing currency.
            if num_has_frac:
                final_num = float(decimal_num)
            else:
                final_num = int(decimal_num)

            try:
                # Convert to words (e.g., 100 -> one hundred)
                num_str = num2words(final_num, **num2words_kwargs)
            except NotImplementedError:
                _LOGGER.exception(
                    "Failed to convert number %s to words for language %s",
                    word.text,
                    word.lang,
                )
                return

            # Add original whitespace back in
            first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
            num_str = first_ws + num_str + last_ws

            # Split into separate words
            for number_word_text in settings.split_words(num_str):
                number_word_text_norm = settings.normalize_whitespace(number_word_text)
                if not number_word_text_norm:
                    continue

                if not settings.keep_whitespace:
                    number_word_text = number_word_text_norm

                number_word = WordNode(
                    node=len(graph),
                    implicit=True,
                    lang=word.lang,
                    text=number_word_text_norm,
                    text_with_ws=number_word_text,
                )
                graph.add_node(number_word.node, data=number_word)
                graph.add_edge(word.node, number_word.node)

    def _verbalize_date(self, graph: GraphType, node: Node):
        """Split dates into words"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if (word.interpret_as != InterpretAs.DATE) or (word.date is None):
            return

        settings = self.get_settings(word.lang)
        assert settings.babel_locale
        assert settings.num2words_lang

        date = word.date
        date_format = word.format or settings.default_date_format

        if "{" not in date_format:
            # Transform into Python format string
            date_format = date_format.strip().upper()

            # MDY -> {M} {D} {Y}
            date_format_str = settings.join_str.join(f"{{{c}}}" for c in date_format)
        else:
            # Assumed to be a Python format string already
            date_format_str = date_format

        day_card_str = ""
        day_ord_str = ""
        month_str = ""
        year_str = ""

        try:
            if ("{M}" in date_format_str) or ("{m}" in date_format_str):
                month_str = babel.dates.format_date(
                    date, "MMMM", locale=settings.babel_locale
                )

            num2words_kwargs = {"lang": settings.num2words_lang}

            if ("{D}" in date_format_str) or ("{d}" in date_format_str):
                # Cardinal day (1 -> one)
                num2words_kwargs["to"] = "cardinal"
                day_card_str = num2words(date.day, **num2words_kwargs)

            if ("{O}" in date_format_str) or ("{o}" in date_format_str):
                # Ordinal day (1 -> first)
                num2words_kwargs["to"] = "ordinal"
                day_ord_str = num2words(date.day, **num2words_kwargs)

            if ("{Y}" in date_format_str) or ("{y}" in date_format_str):
                try:
                    num2words_kwargs["to"] = "year"
                    year_str = num2words(date.year, **num2words_kwargs)
                except Exception:
                    # Fall back to use cardinal number for year
                    num2words_kwargs["to"] = "cardinal"
                    year_str = num2words(date.year, **num2words_kwargs)
        except Exception:
            _LOGGER.exception(
                "Failed to format date %s for language %s", word.text, word.lang
            )
            return

        date_str = date_format_str.format(
            **{
                "M": month_str,
                "m": month_str,
                "D": day_card_str,
                "d": day_card_str,
                "O": day_ord_str,
                "o": day_ord_str,
                "Y": year_str,
                "y": year_str,
            }
        )

        first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
        date_str = first_ws + date_str + last_ws

        # Split into separate words
        for date_word_text in settings.split_words(date_str):
            date_word_text_norm = settings.normalize_whitespace(date_word_text)
            if not date_word_text_norm:
                continue

            if not settings.keep_whitespace:
                date_word_text = date_word_text_norm

            if not date_word_text:
                continue

            date_word = WordNode(
                node=len(graph),
                implicit=True,
                lang=word.lang,
                text=date_word_text_norm,
                text_with_ws=date_word_text,
            )
            graph.add_node(date_word.node, data=date_word)
            graph.add_edge(word.node, date_word.node)

    def _verbalize_time(self, graph: GraphType, node: Node):
        """Split times into words"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if (word.interpret_as != InterpretAs.TIME) or (word.time is None):
            return

        settings = self.get_settings(word.lang)

        if settings.verbalize_time is None:
            # Can't verbalize
            return

        first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
        time_words = list(settings.verbalize_time(word.time))
        last_idx = len(time_words) - 1

        # Split into words
        for word_idx, time_word_text in enumerate(time_words):
            if word_idx == 0:
                time_word_text = first_ws + time_word_text

            if word_idx == last_idx:
                time_word_text += last_ws
            else:
                time_word_text += settings.join_str

            time_word_text_norm = settings.normalize_whitespace(time_word_text)
            if not time_word_text_norm:
                continue

            if not settings.keep_whitespace:
                time_word_text = time_word_text_norm

            if not time_word_text:
                continue

            time_word = WordNode(
                node=len(graph),
                implicit=True,
                lang=word.lang,
                text=time_word_text_norm,
                text_with_ws=time_word_text,
            )

            graph.add_node(time_word.node, data=time_word)
            graph.add_edge(word.node, time_word.node)

            # May contain numbers or initialisms
            self._transform_number(graph, time_word)
            for node_class, node_kwargs in self._split_initialism(graph, time_word):
                new_node = node_class(node=len(graph), **node_kwargs)
                graph.add_node(new_node.node, data=new_node)
                graph.add_edge(time_word.node, new_node.node)

    def _verbalize_currency(
        self, graph: GraphType, node: Node,
    ):
        """Split currency amounts into words"""
        if not isinstance(node, WordNode):
            return

        word = typing.cast(WordNode, node)
        if (
            (word.interpret_as != InterpretAs.CURRENCY)
            or ((word.currency_symbol is None) and (word.currency_name is None))
            or (word.number is None)
        ):
            return

        settings = self.get_settings(word.lang)
        assert settings.num2words_lang

        decimal_num = word.number

        # True if number has non-zero fractional part
        num_has_frac = (decimal_num % 1) != 0

        num2words_kwargs = {"lang": settings.num2words_lang, "to": "currency"}

        # Name of currency (e.g., USD)
        if not word.currency_name:
            currency_name = settings.default_currency
            if settings.currencies:
                # Look up currency in locale
                currency_name = settings.currencies.get(
                    word.currency_symbol or "", settings.default_currency
                )

            word.currency_name = currency_name

        num2words_kwargs["currency"] = word.currency_name

        # Custom separator so we can remove 'zero cents'
        num2words_kwargs["separator"] = "|"

        try:
            num_str = num2words(float(decimal_num), **num2words_kwargs)
        except Exception:
            _LOGGER.exception(
                "Failed to verbalize currency %s for language %s", word, word.lang
            )
            return

        # Post-process currency words
        if num_has_frac:
            # Discard num2words separator
            num_str = num_str.replace("|", "")
        else:
            # Remove 'zero cents' part
            num_str = num_str.split("|", maxsplit=1)[0]

        # Add original whitespace back in
        first_ws, last_ws = settings.get_whitespace(word.text_with_ws)
        num_str = first_ws + num_str + last_ws

        # Split into separate words
        for currency_word_text in settings.split_words(num_str):
            currency_word_text_norm = settings.normalize_whitespace(currency_word_text)
            if not currency_word_text_norm:
                continue

            if not settings.keep_whitespace:
                currency_word_text = currency_word_text_norm

            currency_word = WordNode(
                node=len(graph),
                implicit=True,
                lang=word.lang,
                text=currency_word_text_norm,
                text_with_ws=currency_word_text,
            )
            graph.add_node(currency_word.node, data=currency_word)
            graph.add_edge(word.node, currency_word.node)



"""Utility methods for gruut"""
import itertools
import logging
import os
import re
import ssl
import typing
import xml.etree.ElementTree as etree
from pathlib import Path
from urllib.request import urlopen

import networkx as nx
from gruut_ipa import IPA

from gruut.const import (
    DATA_PROP,
    LANG_ALIASES,
    NODE_TYPE,
    EndElement,
    GraphType,
    InlineLexicon,
    Lexeme,
    Node,
    WordRole,
)
from gruut.resources import _DIR

_LOGGER = logging.getLogger("gruut.utils")

# -----------------------------------------------------------------------------
# Language utilities
# -----------------------------------------------------------------------------

LANG_SPLIT_PATTERN = re.compile(r"[-_]")


def resolve_lang(lang: str) -> str:
    """
    Try to resolve language using aliases.

    Args:
        lang: Language name or alias

    Returns:
        Resolved language name
    """
    lang = lang.lower().replace("_", "-")

    return LANG_ALIASES.get(lang, lang)


def find_lang_dir(
    lang: str,
    search_dirs: typing.Optional[typing.Iterable[typing.Union[str, Path]]] = None,
) -> typing.Optional[Path]:
    """
    Search for a language's model directory by name.

    Tries to find a directory by:

    #. Importing a module name ``gruut_lang_<short_lang>`` where short_lang is "en" for "en-us", etc.
    #. Looking for ``<lang>/lexicon.db`` in each directory in order:

       * ``search_dirs``
       * ``$XDG_CONFIG_HOME/gruut``
       * A "data" directory next to the gruut module

    Args:
        lang: Full language name (e.g., en-us)
        search_dirs: Optional iterable of directory paths to search first

    Returns:
        Path to the language model directory or None if it can't be found
    """
    base_lang = LANG_SPLIT_PATTERN.split(lang)[0].lower()
    lang_module_name = f"gruut_lang_{base_lang}"

    try:
        lang_module = __import__(lang_module_name)

        _LOGGER.debug("(%s) successfully imported %s", lang, lang_module_name)

        return lang_module.get_lang_dir()
    except ImportError:
        _LOGGER.debug("(%s) couldn't import module %s", lang, lang_module_name)
        pass

    search_dirs = typing.cast(typing.List[Path], [Path(p) for p in search_dirs or []])

    # ${XDG_CONFIG_HOME}/gruut or ${HOME}/gruut
    maybe_config_home = os.environ.get("XDG_CONFIG_HOME")
    if maybe_config_home:
        search_dirs.append(Path(maybe_config_home) / "gruut")
    else:
        search_dirs.append(Path.home() / ".config" / "gruut")

    # Data directory *next to* gruut
    search_dirs.append(_DIR.parent / "data")

    _LOGGER.debug("(%s) searching %s for language file(s)", lang, search_dirs)

    for check_dir in search_dirs:
        lang_dir = check_dir / lang
        lexicon_path = lang_dir / "lexicon.db"
        if lexicon_path.is_file():
            _LOGGER.debug("(%s) found language file(s) in %s", lang, lang_dir)
            return lang_dir

    return None


# -----------------------------------------------------------------------------
# Babel
# -----------------------------------------------------------------------------


def get_currency_names(locale_str: str) -> typing.Dict[str, str]:
    """
    Try to get currency names and symbols for a Babel locale.

    Returns:
        Dictionary whose keys are currency symbols (like "$") and whose values are currency names (like "USD")
    """
    currency_names = {}

    try:
        import babel
        import babel.numbers

        locale = babel.Locale(locale_str)
        currency_names = {
            babel.numbers.get_currency_symbol(cn): cn for cn in locale.currency_symbols
        }
    except ImportError:
        # Expected if babel is not installed
        pass
    except Exception:
        _LOGGER.warning("get_currency_names")

    return currency_names


# -----------------------------------------------------------------------------
# Iteration
# -----------------------------------------------------------------------------


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def sliding_window(iterable, n=2):
    """Returns a sliding window of size n over an iterable"""
    iterables = itertools.tee(iterable, n)

    for win_iter, num_skipped in zip(iterables, itertools.count()):
        for _ in range(num_skipped):
            next(win_iter, None)

    return zip(*iterables)


# -----------------------------------------------------------------------------
# XML
# -----------------------------------------------------------------------------

NO_NAMESPACE_PATTERN = re.compile(r"^{[^}]+}")


def tag_no_namespace(tag: str) -> str:
    """Remove namespace from XML tag"""
    return NO_NAMESPACE_PATTERN.sub("", tag)


def attrib_no_namespace(
    element: etree.Element, name: str, default: typing.Any = None
) -> typing.Any:
    """Search for an attribute by key without namespaces"""
    for key, value in element.attrib.items():
        key_no_ns = NO_NAMESPACE_PATTERN.sub("", key)
        if key_no_ns == name:
            return value

    return default


def text_and_elements(element, is_last=False):
    """Yields element, text, sub-elements, end element, and tail"""
    element_metadata = None

    if is_last:
        # True if this is the last child element of a parent.
        # Used to preserve whitespace.
        element_metadata = {"is_last": True}

    yield element, element_metadata

    # Text before any tags (or end tag)
    text = element.text if element.text is not None else ""
    if text.strip():
        yield text

    children = list(element)
    last_child_idx = len(children) - 1

    for child_idx, child in enumerate(children):
        # Sub-elements
        is_last = child_idx == last_child_idx
        yield from text_and_elements(child, is_last=is_last)

    # End of current element
    yield EndElement(element)

    # Text after the current tag
    tail = element.tail if element.tail is not None else ""
    if tail.strip():
        yield tail


def load_lexicon(
    uri: str,
    lexicon: InlineLexicon,
    ssl_context: typing.Optional[ssl.SSLContext] = None,
):
    """Loads a pronunciation lexicon from a URI"""
    if ssl_context is None:
        ssl_context = ssl.create_default_context()

    with urlopen(uri, context=ssl_context) as response:
        tree = etree.parse(response)
        for lexeme_elem in tree.getroot():
            if tag_no_namespace(lexeme_elem.tag) != "lexeme":
                continue

            lexeme = Lexeme()

            role_str = attrib_no_namespace(lexeme_elem, "role")
            if role_str:
                lexeme.roles = set(role_str.strip().split())

            for lexeme_child in lexeme_elem:

                child_tag = tag_no_namespace(lexeme_child.tag)
                if child_tag == "grapheme":
                    if lexeme_child.text:
                        lexeme.grapheme = lexeme_child.text.strip()
                elif child_tag == "phoneme":
                    if lexeme_child.text:
                        lexeme.phonemes = maybe_split_ipa(lexeme_child.text.strip())

            if lexeme.grapheme and lexeme.phonemes:
                role_phonemes = lexicon.words.get(lexeme.grapheme)
                if role_phonemes is None:
                    role_phonemes = {}
                    lexicon.words[lexeme.grapheme] = role_phonemes

                assert role_phonemes is not None

                roles = lexeme.roles or [WordRole.DEFAULT]
                for role in roles:
                    role_phonemes[role] = lexeme.phonemes


# -----------------------------------------------------------------------------
# Text
# -----------------------------------------------------------------------------

NON_WORDS_PATTERN = re.compile(r"\W")


def remove_non_word_chars(s: str) -> str:
    """Removes non-word characters from a string"""
    return NON_WORDS_PATTERN.sub("", s)


def maybe_split_ipa(s: str) -> typing.List[str]:
    """Split on whitespace if a space is present, otherwise return string as list of graphemes"""
    if " " in s:
        # Manual separation
        return s.split()

    # Automatic separation
    return IPA.graphemes(s)


# -----------------------------------------------------------------------------
# Graph
# -----------------------------------------------------------------------------


def print_graph(
    graph: GraphType,
    node: typing.Union[NODE_TYPE, Node],
    indent: str = "--",
    level: int = 1,
    print_func=print,
):
    """Prints a graph to the console"""
    if isinstance(node, Node):
        n_data = node
        graph_node = node.node
    else:
        graph_node = node
        n_data = typing.cast(Node, graph.nodes[graph_node][DATA_PROP])

    print_func(indent * level, graph_node, n_data)
    for succ_node in graph.successors(graph_node):
        print_graph(
            graph, succ_node, indent=indent, level=level + 1, print_func=print_func
        )


def leaves(graph: GraphType, node: Node):
    """Iterate through the leaves of a graph in depth-first order"""
    for dfs_node in nx.dfs_preorder_nodes(graph, node.node):
        if not graph.out_degree(dfs_node) == 0:
            continue

        yield graph.nodes[dfs_node][DATA_PROP]


def pipeline_split(split_func, graph: GraphType, parent_node: Node,) -> bool:
    """Splits leaf nodes of tree into zero or more sub-nodes"""
    was_changed = False

    for leaf_node in list(leaves(graph, parent_node)):
        for node_class, node_kwargs in split_func(graph, leaf_node):
            new_node = node_class(node=len(graph), **node_kwargs)
            graph.add_node(new_node.node, data=new_node)
            graph.add_edge(leaf_node.node, new_node.node)
            was_changed = True

    return was_changed


def pipeline_transform(transform_func, graph: GraphType, parent_node: Node,) -> bool:
    """Transforms leaves of tree with a custom function"""
    was_changed = False

    for leaf_node in list(leaves(graph, parent_node)):
        if transform_func(graph, leaf_node):
            was_changed = True

    return was_changed



#!/usr/bin/env python3
"""Tests for GraphemesToPhonemes class"""
import unittest

from gruut.g2p import GraphemesToPhonemes


class GraphemesToPhonemesTestCase(unittest.TestCase):
    """Test cases for GraphemesToPhonemes class"""

    def test_encode_decode(self):
        """Test encode/decode functions for pycrfsuite features"""
        s = "ði ıntəˈnæʃənəl fəˈnɛtık əsoʊsiˈeıʃn"
        self.assertEqual(
            GraphemesToPhonemes.decode_string(GraphemesToPhonemes.encode_string(s)), s
        )

    def test_features(self):
        """Test word features"""
        word = "test"
        actual_features = GraphemesToPhonemes.word2features(
            word, chars_forward=1, chars_backward=1, encode=False
        )

        expected_features = [
            {"bias": 1.0, "grapheme": "t", "begin": True, "grapheme+1": "e"},
            {"bias": 1.0, "grapheme": "e", "grapheme+1": "s", "grapheme-1": "t"},
            {"bias": 1.0, "grapheme": "s", "grapheme+1": "t", "grapheme-1": "e"},
            {"bias": 1.0, "grapheme": "t", "end": True, "grapheme-1": "s"},
        ]

        self.assertEqual(expected_features, actual_features)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Tests for PartOfSpeechTagger class"""
import copy
import unittest

from gruut.pos import PartOfSpeechTagger


class PartOfSpeechTaggerTestCase(unittest.TestCase):
    """Test cases for PartOfSpeechTagger class"""

    def test_encode_decode(self):
        """Test encode/decode functions for pycrfsuite features"""
        s = "ði ıntəˈnæʃənəl fəˈnɛtık əsoʊsiˈeıʃn"
        self.assertEqual(
            PartOfSpeechTagger.decode_string(PartOfSpeechTagger.encode_string(s)), s
        )

    def test_features(self):
        """Test sentence features"""
        sentence = "1 test .".split()
        word_features = {
            "1": {
                "bias": 1.0,
                "word": "1",
                "len(word)": 1,
                "word.ispunctuation": False,
                "word.isdigit()": True,
                "word[:2]": "1",
                "word[-2:]": "1",
            },
            "test": {
                "bias": 1.0,
                "word": "test",
                "len(word)": 4,
                "word.ispunctuation": False,
                "word[-2:]": "st",
                "word[:2]": "te",
                "word.isdigit()": False,
            },
            ".": {
                "bias": 1.0,
                "word": ".",
                "len(word)": 1,
                "word.ispunctuation": True,
                "word.isdigit()": False,
                "word[-2:]": ".",
                "word[:2]": ".",
            },
        }

        def add_prefix(d, prefix):
            return {f"{prefix}{k}": v for k, v in d.items()}

        # Add context
        context_features = copy.deepcopy(word_features)
        context_features["1"].update(add_prefix(word_features["test"], "+1:"))

        context_features["test"].update(add_prefix(word_features["1"], "-1:"))
        context_features["test"].update(add_prefix(word_features["."], "+1:"))

        context_features["."].update(add_prefix(word_features["test"], "-1:"))

        # Add BOS/EOS
        context_features["1"]["BOS"] = True
        context_features["."]["EOS"] = True

        expected_features = [
            context_features["1"],
            context_features["test"],
            context_features["."],
        ]

        actual_features = PartOfSpeechTagger.sent2features(
            sentence,
            words_forward=1,
            words_backward=1,
            chars_front=2,
            chars_back=2,
            encode=False,
        )

        self.assertEqual(expected_features, actual_features)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Tests for English class"""
import unittest

from gruut import sentences


class EnglishTestCase(unittest.TestCase):
    """Test cases for English"""

    def test_unclean_text(self):
        """Test text with lots of noise"""
        text = (
            "IT’S <a> 'test' (seNtEnce) for-only $100, Dr., & [I] ## *like* ## it 100%!"
        )
        sentence = next(sentences(text, lang="en_US"))

        self.assertEqual(
            [
                "IT'S",
                "<",
                "a",
                ">",
                "'",
                "test",
                "'",
                "(",
                "seNtEnce",
                ")",
                "for",
                "only",
                "one",
                "hundred",
                "dollars",
                ",",
                "Doctor",
                ",",
                "and",
                "[",
                "I",
                "]",
                "*",
                "like",
                "*",
                "it",
                "one",
                "hundred",
                "percent",
                "!",
            ],
            [word.text for word in sentence],
        )

    def test_spell_out(self):
        """Test spell-out in say-as SSML tag"""
        text = '<say-as interpret-as="spell-out">abc@1+2-3*.*</say-as>'
        sentence = next(sentences(text, lang="en_US", ssml=True))

        self.assertEqual(
            [
                "a",
                "b",
                "c",
                "at",
                "one",
                "plus",
                "two",
                "dash",
                "three",
                "star",
                "dot",
                "star",
            ],
            [word.text for word in sentence],
        )

    def test_initialisms(self):
        """Test expansion of initialisms"""
        text = "ABC abc A.B.C."
        sentence = next(sentences(text, lang="en_US"))

        self.assertEqual(
            ["A", "B", "C", "abc", "A", "B", "C"], [word.text for word in sentence],
        )

    def test_dates(self):
        """Test expansion of dates"""
        text = "1/4/1999 vs. 4/1/1999"
        sentence = next(sentences(text, lang="en_US"))

        self.assertEqual(
            [
                "January",
                "fourth",
                ",",
                "nineteen",
                "ninety",
                "nine",
                "versus",
                "April",
                "first",
                ",",
                "nineteen",
                "ninety",
                "nine",
            ],
            [word.text for word in sentence],
        )

    def test_ordinals(self):
        """Test parsing of ordinal numbers"""
        text = "1st, 2nd, 3rd, 4th, 5th, 23rd, 32nd, 44th, 121st, 5,111st."
        sentence = next(sentences(text, lang="en_US"))

        self.assertEqual(
            [
                "first",
                ",",
                "second",
                ",",
                "third",
                ",",
                "fourth",
                ",",
                "fifth",
                ",",
                "twenty",
                "third",
                ",",
                "thirty",
                "second",
                ",",
                "forty",
                "fourth",
                ",",
                "one",
                "hundred",
                "and",
                "twenty",
                "first",
                ",",
                "five",
                "thousand",
                ",",
                "one",
                "hundred",
                "and",
                "eleventh",
                ".",
            ],
            [word.text for word in sentence],
        )


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Converts a text lexicon to a gruut sqlite3 database"""
import argparse
import sqlite3
import sys

# -----------------------------------------------------------------------------


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(prog="lexicon2db.py")
    parser.add_argument(
        "--casing",
        required=True,
        choices=("keep", "lower", "upper"),
        help="Casing to apply to words",
    )
    parser.add_argument(
        "--lexicon",
        required=True,
        help="Text lexicon to read with <WORD> <PHONEME> <PHONEME> ...",
    )
    parser.add_argument("--database", required=True, help="SQLite database to write")
    parser.add_argument(
        "--role", action="store_true", help="Lexicon includes word roles (2nd column)",
    )
    parser.add_argument(
        "--empty-role",
        default="_",
        help="String used to identify empty word role (see --role)",
    )
    args = parser.parse_args()

    # -------------------------------------------------------------------------

    word_casing = None

    if args.casing == "lower":
        word_casing = str.lower
    elif args.casing == "upper":
        word_casing = str.upper

    # -------------------------------------------------------------------------

    conn = sqlite3.connect(args.database)

    # Re-create tables in output
    conn.execute("DROP TABLE IF EXISTS word_phonemes")
    conn.execute("DROP TABLE IF EXISTS g2p_alignments")
    conn.execute(
        "CREATE TABLE IF NOT EXISTS word_phonemes "
        + "(id INTEGER PRIMARY KEY AUTOINCREMENT, word TEXT, pron_order INTEGER, phonemes TEXT, role TEXT);"
    )
    conn.execute(
        "CREATE TABLE IF NOT EXISTS g2p_alignments "
        + "(id INTEGER PRIMARY KEY AUTOINCREMENT, word TEXT, alignment TEXT);"
    )
    conn.commit()

    # word -> pron_order
    pron_orders = {}

    if args.lexicon == "-":
        lexicon_file = sys.stdin
    else:
        lexicon_file = open(args.lexicon, "r", encoding="utf-8")

    with lexicon_file, conn:
        for i, line in enumerate(lexicon_file):
            try:
                line = line.strip()
                if (not line) or line.startswith(";") or (" " not in line):
                    # Skip blank lines and comments.
                    # Also skip lines without a pronunciation.
                    continue

                role = ""

                if args.role:
                    # With role
                    word, role, phonemes_str = line.split(maxsplit=2)

                    if role == args.empty_role:
                        role = ""
                    elif ":" not in role:
                        role = f"gruut:{role}"
                else:
                    # Without part of speech
                    word, phonemes_str = line.split(maxsplit=1)

                if word_casing:
                    word = word_casing(word)

                pron_order = pron_orders.get(word, 0)

                # Don't commit on every word, or it will be terribly slow
                conn.execute(
                    "INSERT into word_phonemes (word, pron_order, phonemes, role) VALUES (?, ?, ?, ?)",
                    (word, pron_order, phonemes_str, role),
                )

                pron_orders[word] = pron_order + 1
            except Exception as e:
                print("Error on line", i + 1, "-", line)
                raise e


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Part of speech tagging using python CRF suite.

Credit to: https://towardsdatascience.com/pos-tagging-using-crfs-ea430c5fb78b

Training requires conllu package:

.. code-block:: sh

    pip install conllu

Training data comes from Univeral Dependencies (https://universaldependencies.org/)

Example:

.. code-block:: sh

    python3 -m gruut.pos train --conllu train.conllu --output model.crf --label xpos

Pre-trained models have the following settings:

* c1 = 0.25
* c2 = 0.3
* max-iterations = 100

English model is trained with "xpos" label.
French model is trained with "upos" label.
"""
import argparse
import base64
import logging
import os
import string
import sys
import time
import typing
from pathlib import Path

import jsonlines
import pycrfsuite

_LOGGER = logging.getLogger("gruut.pos")

# -----------------------------------------------------------------------------

FEATURES_TYPE = typing.Dict[
    str, typing.Union[str, bool, int, float, typing.Sequence[str]]
]


class PartOfSpeechTagger:
    """Part of speech tagger using a pre-trained CRF model"""

    def __init__(
        self, crf_tagger: typing.Union[str, Path, pycrfsuite.Tagger], **kwargs
    ):
        if isinstance(crf_tagger, pycrfsuite.Tagger):
            self.crf_tagger = crf_tagger
        else:
            # Load model
            self.crf_tagger = pycrfsuite.Tagger()
            self.crf_tagger.open(str(crf_tagger))

    def __call__(self, words: typing.Sequence[str]) -> typing.Sequence[str]:
        """Returns POS tag for each word"""
        features = PartOfSpeechTagger.sent2features(words)
        return self.crf_tagger.tag(features)

    @staticmethod
    def local_features(
        word: str,
        prefix: str = "",
        bias: float = 1.0,
        add_punctuation: bool = True,
        add_digit: bool = True,
        add_length: bool = True,
        chars_front: int = 3,
        chars_back: int = 3,
        encode: bool = True,
    ) -> FEATURES_TYPE:
        """Get features for a single word"""
        features: FEATURES_TYPE = {
            f"{prefix}bias": bias,
            f"{prefix}word": PartOfSpeechTagger.encode_string(word) if encode else word,
        }

        if add_length:
            features[f"{prefix}len(word)"] = len(word)

        if add_punctuation:
            features[f"{prefix}word.ispunctuation"] = word in string.punctuation

        if add_digit:
            features[f"{prefix}word.isdigit()"] = word.isdigit()

        # Chunks from front
        for i in range(2, chars_front + 1):
            features[f"{prefix}word[:{i}]"] = word[:i]

        # Chunks from pack
        for i in range(2, chars_back + 1):
            features[f"{prefix}word[-{i}:]"] = word[-i:]

        return features

    @staticmethod
    def word2features(
        sentence: typing.Sequence[str],
        i: int,
        add_bos: bool = True,
        add_eos: bool = True,
        words_backward: int = 2,
        words_forward: int = 2,
        **kwargs,
    ) -> FEATURES_TYPE:
        """Get features for a word and surrounding context"""
        word = sentence[i]
        num_words = len(sentence)
        features = PartOfSpeechTagger.local_features(word, **kwargs)

        if (i == 0) and add_bos:
            features["BOS"] = True

        if (i == (num_words - 1)) and add_eos:
            features["EOS"] = True

        for j in range(1, words_backward + 1):
            if i >= j:
                word_prev = sentence[i - j]
                features.update(
                    PartOfSpeechTagger.local_features(
                        word_prev, prefix=f"-{j}:", **kwargs
                    )
                )

        for j in range(1, words_forward + 1):
            if i < (num_words - j):
                word_next = sentence[i + j]
                features.update(
                    PartOfSpeechTagger.local_features(
                        word_next, prefix=f"+{j}:", **kwargs
                    )
                )

        return features

    @staticmethod
    def sent2features(
        sentence: typing.Sequence[str], **kwargs
    ) -> typing.List[FEATURES_TYPE]:
        """Get features for all words in a sentence"""
        return [
            PartOfSpeechTagger.word2features(sentence, i, **kwargs)
            for i in range(len(sentence))
        ]

    @staticmethod
    def encode_string(s: str) -> str:
        """Encodes string in a form that crfsuite will accept (ASCII) and can be decoded"""
        return base64.b64encode(s.encode()).decode("ascii")

    @staticmethod
    def decode_string(s: str) -> str:
        """Decodes a string encoded by encode_string"""
        return base64.b64decode(s.encode("ascii")).decode()


# -----------------------------------------------------------------------------


def train_model(
    conllu_path: typing.Union[str, Path],
    output_path: typing.Union[str, Path],
    label: str = "xpos",
    c1: float = 0.25,
    c2: float = 0.3,
    max_iterations: int = 100,
):
    """Train a new model from CONLLU data"""
    try:
        import conllu
    except ImportError as e:
        _LOGGER.critical("conllu package is required for training")
        _LOGGER.critical("pip install 'conllu>=4.4'")
        raise e

    conllu_path = Path(conllu_path)
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    _LOGGER.debug("Loading train file (%s)", conllu_path)
    with open(conllu_path, "r", encoding="utf-8") as conllu_file:
        train_sents = conllu.parse(conllu_file.read())

    _LOGGER.debug("Training model for %s max iteration(s)", max_iterations)
    trainer = pycrfsuite.Trainer(verbose=False)

    _LOGGER.debug("Getting features for %s training sentence(s)", len(train_sents))
    for sent in train_sents:
        words = [token["form"] for token in sent]
        features = PartOfSpeechTagger.sent2features(words)

        labels = []
        skip_sent = False
        for token in sent:
            token_label = token.get(label)
            if token_label is None:
                _LOGGER.warning("Example has empty label for %s: %s", token, sent)
                skip_sent = True
                break

            labels.append(token_label)

        if skip_sent:
            continue

        trainer.append(features, labels)

    trainer.set_params(
        {
            "c1": c1,  # coefficient for L1 penalty
            "c2": c2,  # coefficient for L2 penalty
            "max_iterations": max_iterations,  # stop earlier
            # include transitions that are possible, but not observed
            "feature.possible_transitions": True,
        }
    )
    _LOGGER.debug(trainer.get_params())

    # Begin training
    _LOGGER.info("Training")

    start_time = time.perf_counter()
    trainer.train(str(output_path))
    end_time = time.perf_counter()

    _LOGGER.info("Training completed in %s second(s)", end_time - start_time)
    _LOGGER.info(trainer.logparser.last_iteration)


def do_train(args):
    """CLI method for train_model"""
    train_model(
        conllu_path=args.conllu,
        output_path=args.output,
        label=args.label,
        c1=args.c1,
        c2=args.c2,
        max_iterations=args.max_iterations,
    )


# -----------------------------------------------------------------------------


def do_print_labels(args):
    """Print label set from a CONLLU file"""
    try:
        import conllu
    except ImportError as e:
        _LOGGER.critical("conllu package is required for training")
        _LOGGER.critical("pip install 'conllu>=4.4'")
        raise e

    labels = set()
    with open(args.conllu, "r", encoding="utf-8") as conllu_file:
        for sent in conllu.parse(conllu_file.read()):
            for token in sent:
                token_label = token.get(args.label)
                if token_label is not None:
                    labels.add(token_label)

    print(sorted(labels))


# -----------------------------------------------------------------------------


def do_predict(args):
    """CLI method for predict"""
    tagger = PartOfSpeechTagger(args.model)

    if args.texts:
        lines = args.texts
    else:
        lines = sys.stdin

        if os.isatty(sys.stdin.fileno()):
            print("Reading sentences from stdin...", file=sys.stderr)

    writer = jsonlines.Writer(sys.stdout, flush=True)
    for line in lines:
        line = line.strip()
        if not line:
            continue

        words = line.split()
        words_and_tags = list(zip(words, tagger(words)))

        writer.write(words_and_tags)


def do_test(args):
    """CLI method for testing"""
    try:
        import conllu
    except ImportError as e:
        _LOGGER.critical("conllu package is required for training")
        _LOGGER.critical("pip install 'conllu>=4.4'")
        raise e

    tagger = PartOfSpeechTagger(args.model)

    _LOGGER.debug("Testing file (%s)", args.conllu)

    num_sentences = 0
    num_words = 0
    sents_with_errors = 0
    total_errors = 0
    with open(args.conllu, "r", encoding="utf-8") as conllu_file:
        for sent in conllu.parse(conllu_file.read()):
            words = [token["form"] for token in sent]
            actual_labels = [token.get(args.label) for token in sent]
            expected_labels = tagger(words)

            had_error = False
            for actual, expected in zip(actual_labels, expected_labels):
                if actual != expected:
                    total_errors += 1
                    had_error = True

                num_words += 1

            if had_error:
                sents_with_errors += 1

            num_sentences += 1

    if (num_sentences < 1) or (num_words < 1):
        return

    print(
        "{0} out of {1} word(s) had an incorrect tag ({2:0.2f}%)".format(
            total_errors, num_words, total_errors / num_words
        )
    )
    print(
        "{0} out of {1} sentence(s) had at least one error ({2:0.2f}%)".format(
            sents_with_errors, num_sentences, sents_with_errors / num_sentences
        )
    )


# -----------------------------------------------------------------------------


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(prog="pos.py")

    # Create subparsers for each sub-command
    sub_parsers = parser.add_subparsers()
    sub_parsers.required = True
    sub_parsers.dest = "command"

    # -----
    # Train
    # -----
    train_parser = sub_parsers.add_parser(
        "train", help="Train a new POS model from CONLLU file"
    )
    train_parser.add_argument(
        "--conllu", required=True, help="CONLLU file with training data"
    )
    train_parser.add_argument(
        "--output", required=True, help="Path to write output model"
    )
    train_parser.add_argument(
        "--label", default="xpos", help="Field to predict in training data"
    )
    train_parser.add_argument("--c1", type=float, default=0.25, help="L1 penalty")
    train_parser.add_argument("--c2", type=float, default=0.3, help="L2 penalty")
    train_parser.add_argument(
        "--max-iterations",
        type=int,
        default=100,
        help="Maximum number of iterations to train for",
    )
    train_parser.set_defaults(func=do_train)

    # ----
    # Test
    # ----
    test_parser = sub_parsers.add_parser(
        "test", help="Test a POS model on a CONLLU file"
    )
    test_parser.add_argument("--model", required=True, help="Path to POS tagger model")
    test_parser.add_argument(
        "--conllu", required=True, help="CONLLU file with testing data"
    )
    test_parser.add_argument(
        "--label", default="xpos", help="Field to predict in training data"
    )
    test_parser.set_defaults(func=do_test)

    # ------------
    # Print Labels
    # ------------
    print_labels_parser = sub_parsers.add_parser(
        "print-labels", help="Print set of unique labels from a CONLLU file"
    )
    print_labels_parser.add_argument(
        "--conllu", required=True, help="CONLLU file with training data"
    )
    print_labels_parser.add_argument(
        "--label", default="xpos", help="Field to predict in training data"
    )
    print_labels_parser.set_defaults(func=do_print_labels)

    # -------
    # Predict
    # -------
    predict_parser = sub_parsers.add_parser(
        "predict", help="Predict POS tags for sentence(s)"
    )
    predict_parser.add_argument(
        "--model", required=True, help="Path to POS tagger model"
    )
    predict_parser.add_argument("texts", nargs="*", help="Sentences")
    predict_parser.set_defaults(func=do_predict)

    # ----------------
    # Shared arguments
    # ----------------
    for sub_parser in [train_parser, predict_parser, test_parser, print_labels_parser]:
        sub_parser.add_argument(
            "--debug", action="store_true", help="Print DEBUG messages to console"
        )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    args.func(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Command-line interface to gruut"""
import argparse
import csv
import dataclasses
import logging
import os
import sys
from enum import Enum
from pathlib import Path

import jsonlines

from gruut.const import KNOWN_LANGS
from gruut.text_processor import TextProcessor
from gruut.utils import print_graph

# -----------------------------------------------------------------------------

_LOGGER = logging.getLogger("gruut")

# Path to gruut base directory
_DIR = Path(__file__).parent


class StdinFormat(str, Enum):
    """Format of standard input"""

    AUTO = "auto"
    """Choose based on SSML state"""

    LINES = "lines"
    """Each line is a separate sentence/document"""

    DOCUMENT = "document"
    """Entire input is one document"""


# -----------------------------------------------------------------------------


def main():
    """Main entry point"""
    if len(sys.argv) < 2:
        # Print known languages and exit
        print("Languages:", *sorted(list(KNOWN_LANGS)))
        sys.exit(0)
    elif sys.argv[1] == "--version":
        from gruut import __version__

        print(__version__)
        sys.exit(0)

    args = get_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    if args.espeak:
        args.model_prefix = "espeak"

    # -------------------------------------------------------------------------

    text_processor = TextProcessor(
        default_lang=args.language, model_prefix=args.model_prefix,
    )

    if args.debug:
        _LOGGER.debug(text_processor.settings)

    if args.text:
        # Use arguments
        lines = args.text
    else:
        # Use stdin
        stdin_format = StdinFormat.LINES

        if (args.stdin_format == StdinFormat.AUTO) and args.ssml:
            # Assume SSML input is entire document
            stdin_format = StdinFormat.DOCUMENT

        if stdin_format == StdinFormat.DOCUMENT:
            # One big line
            lines = [sys.stdin.read()]
        else:
            # Multiple lines
            lines = sys.stdin

        if os.isatty(sys.stdin.fileno()):
            print("Reading input from stdin...", file=sys.stderr)

    if args.csv:
        writer = csv.writer(sys.stdout, delimiter=args.csv_delimiter)

        def input_text(lines):
            reader = csv.reader(lines, delimiter=args.csv_delimiter)
            for row in reader:
                text = row[1]
                yield (text, row)

        def output_sentences(sentences, writer, text_data=None):
            row = list(text_data)
            row.append(
                args.sentence_separator.join(
                    args.word_separator.join(w.text for w in sentence if w.is_spoken)
                    for sentence in sentences
                )
            )

            phonemes = [
                args.phoneme_separator.join(w.phonemes)
                for sentence in sentences
                for w in sentence
                if w.phonemes
            ]

            row.append(args.phoneme_word_separator.join(phonemes))
            writer.writerow(row)

    else:
        writer = jsonlines.Writer(sys.stdout, flush=True)

        def input_text(lines):
            for line in lines:
                yield (line, None)

        def output_sentences(sentences, writer, text_data=None):
            for sentence in sentences:
                sentence_dict = dataclasses.asdict(sentence)
                writer.write(sentence_dict)

    for text, text_data in input_text(lines):
        try:
            graph, root = text_processor(
                text,
                ssml=args.ssml,
                pos=(not args.no_pos),
                phonemize=(not (args.no_lexicon and args.no_g2p)),
                post_process=(not args.no_post_process),
                verbalize_numbers=(not args.no_numbers),
                verbalize_currency=(not args.no_currency),
                verbalize_dates=(not args.no_dates),
                verbalize_times=(not args.no_times),
            )

            if args.debug:
                print_graph(
                    graph,
                    root,
                    print_func=lambda *print_args: _LOGGER.debug(
                        " ".join(str(a) for a in print_args)
                    ),
                )

            # Output sentences
            sentences = list(
                text_processor.sentences(
                    graph,
                    root,
                    major_breaks=(not args.no_major_breaks),
                    minor_breaks=(not args.no_minor_breaks),
                    punctuations=(not args.no_punctuation),
                )
            )

            output_sentences(sentences, writer, text_data)
        except Exception as e:
            _LOGGER.exception(text)

            if not args.no_fail:
                raise TextProcessingError(text) from e


# -----------------------------------------------------------------------------


class TextProcessingError(Exception):
    """Raised when a line of input results in an exception"""

    pass


# -----------------------------------------------------------------------------


def get_args() -> argparse.Namespace:
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(prog="gruut")
    parser.add_argument(
        "-l",
        "--language",
        default="en-us",
        help="Default language code (default: en-us)",
    )

    parser.add_argument("text", nargs="*", help="Text to tokenize (default: stdin)")
    parser.add_argument(
        "--ssml", action="store_true", help="Input text is SSML",
    )
    parser.add_argument(
        "--stdin-format",
        choices=[str(v.value) for v in StdinFormat],
        default=StdinFormat.AUTO,
        help="Format of stdin text (default: auto)",
    )

    # Disable features
    parser.add_argument(
        "--no-numbers",
        action="store_true",
        help="Disable number replacement (1 -> one)",
    )
    parser.add_argument(
        "--no-currency",
        action="store_true",
        help="Disable currency replacement ($1 -> one dollar)",
    )
    parser.add_argument(
        "--no-dates",
        action="store_true",
        help="Disable date replacement (4/1/2021 -> April first twenty twenty one)",
    )
    parser.add_argument(
        "--no-times",
        action="store_true",
        help="Disable time replacement (4:01pm -> four oh one P M)",
    )
    parser.add_argument(
        "--no-pos", action="store_true", help="Disable part of speech tagger",
    )
    parser.add_argument(
        "--no-lexicon", action="store_true", help="Disable phoneme lexicon database",
    )
    parser.add_argument(
        "--no-g2p", action="store_true", help="Disable grapheme to phoneme guesser",
    )
    parser.add_argument(
        "--no-punctuation",
        action="store_true",
        help="Don't output punctuations (quotes, brackets, etc.)",
    )
    parser.add_argument(
        "--no-major-breaks",
        action="store_true",
        help="Don't output major breaks (periods, question marks, etc.)",
    )
    parser.add_argument(
        "--no-minor-breaks",
        action="store_true",
        help="Don't output minor breaks (commas, semicolons, etc.)",
    )
    parser.add_argument(
        "--no-post-process",
        action="store_true",
        help="Disable post-processing of sentences (e.g., liasons)",
    )
    parser.add_argument(
        "--no-fail", action="store_true", help="Skip lines that result in errors",
    )

    # Miscellaneous
    parser.add_argument(
        "--espeak",
        action="store_true",
        help="Use eSpeak versions of lexicons (overrides --model-prefix)",
    )
    parser.add_argument(
        "--model-prefix",
        help="Sub-directory of gruut language data files with different lexicon, etc. (e.g., espeak)",
    )
    parser.add_argument(
        "--csv", action="store_true", help="Input text is id|text (see --csv-delimiter)"
    )
    parser.add_argument(
        "--csv-delimiter", default="|", help="Delimiter for input text with --csv"
    )
    parser.add_argument(
        "--sentence-separator",
        default=". ",
        help="String used to separate sentences in CSV output",
    )
    parser.add_argument(
        "--word-separator",
        default=" ",
        help="String used to separate words in CSV output",
    )
    parser.add_argument(
        "--phoneme-word-separator",
        default="#",
        help="String used to separate phonemes in CSV output",
    )
    parser.add_argument(
        "--phoneme-separator",
        default=" ",
        help="String used to separate words in CSV output phonemes",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to console"
    )

    return parser.parse_args()


# -----------------------------------------------------------------------------


if __name__ == "__main__":
    main()
